"""Convert OASIS dataset (https://sites.wustl.edu/oasisbrains/) to BIDS."""

from pathlib import Path
from typing import Optional

import nibabel as nb
import numpy as np

from clinica.iotools.abstract_converter import Converter
from clinica.utils.filemanip import UserProvidedPath

__all__ = ["convert"]


def convert(
    path_to_dataset: UserProvidedPath,
    bids_dir: UserProvidedPath,
    path_to_clinical: UserProvidedPath,
    subjects: Optional[UserProvidedPath] = None,
    n_procs: Optional[int] = 1,
    **kwargs,
):
    from clinica.iotools.bids_utils import StudyName
    from clinica.iotools.converters.factory import get_converter_name
    from clinica.utils.stream import cprint

    from ..utils import validate_input_path

    path_to_dataset = validate_input_path(path_to_dataset)
    bids_dir = validate_input_path(bids_dir, check_exist=False)
    path_to_clinical = validate_input_path(path_to_clinical)
    if subjects:
        cprint(
            (
                f"Subject filtering is not yet implemented in {get_converter_name(StudyName.OASIS)} converter. "
                "All subjects available will be converted."
            ),
            lvl="warning",
        )
    OasisToBids().convert(
        path_to_dataset,
        bids_dir,
        path_to_clinical,
        n_procs=n_procs,
    )


class OasisToBids(Converter):
    def convert(
        self,
        source_dir: Path,
        destination_dir: Path,
        clinical_data_dir: Path,
        n_procs: Optional[int] = 1,
    ):
        self.convert_images(source_dir, destination_dir, n_procs=n_procs)
        self.convert_clinical_data(clinical_data_dir, destination_dir)

    def convert_clinical_data(self, clinical_data_dir: Path, bids_dir: Path):
        """Convert the clinical data defined inside the clinical_specifications.xlx into BIDS.

        Args:
            clinical_data_dir: path to the folder with the original clinical data
            bids_dir: path to the BIDS directory
        """
        from clinica.iotools.bids_utils import get_bids_subjs_list
        from clinica.utils.stream import cprint

        cprint("Converting clinical data...", lvl="info")
        bids_ids = get_bids_subjs_list(bids_dir)
        self._create_participants_tsv(clinical_data_dir, bids_dir, bids_ids)
        sessions = self._create_sessions_tsv(clinical_data_dir, bids_dir, bids_ids)
        self._create_scans_tsv(clinical_data_dir, bids_dir, bids_ids, sessions)
        self._create_modality_agnostic_files(bids_dir)

    def _create_participants_tsv(
        self,
        clinical_data_dir: Path,
        bids_dir: Path,
        bids_ids: list[str],
    ):
        from clinica.iotools.bids_utils import StudyName, create_participants_df

        participants_df = create_participants_df(
            study_name=StudyName.OASIS,
            clinical_specifications_folder=Path(__file__).parents[1] / "specifications",
            clinical_data_dir=clinical_data_dir,
            bids_ids=bids_ids,
        )
        # Replace the values of the diagnosis_bl column
        participants_df["diagnosis_bl"].replace([0.0, np.nan], "CN", inplace=True)
        participants_df["diagnosis_bl"].replace(
            [0.5, 1.0, 1.5, 2.0], "AD", inplace=True
        )
        # Following line has no sense
        # participants_df['diagnosis_bl'].replace(participants_df['diagnosis_bl']>0.0, 'AD', inplace=True)
        participants_df = participants_df.fillna("n/a")
        participants_df.to_csv(
            bids_dir / "participants.tsv",
            sep="\t",
            index=False,
            encoding="utf-8",
        )

    def _create_sessions_tsv(
        self,
        clinical_data_dir: Path,
        bids_dir: Path,
        bids_ids: list[str],
    ) -> dict:
        from .oasis_to_bids_utils import create_sessions_dict, write_sessions_tsv

        sessions_dict = create_sessions_dict(
            clinical_data_dir=clinical_data_dir,
            bids_dir=bids_dir,
            clinical_specifications_folder=Path(__file__).parents[1] / "specifications",
            bids_ids=bids_ids,
        )

        # todo : when tested add to create_sessions_dict bc specific to oasis1
        for bids_id in bids_ids:
            sessions_dict[bids_id]["M000"]["diagnosis"] = (
                "AD" if sessions_dict[bids_id]["M000"]["diagnosis"] > 0 else "CN"
            )
        write_sessions_tsv(bids_dir, sessions_dict)

        return sessions_dict

    def _create_scans_tsv(
        self,
        clinical_data_dir: Path,
        bids_dir: Path,
        bids_ids: list[str],
        sessions: dict,
    ):
        from clinica.iotools.bids_utils import (
            StudyName,
            create_scans_dict,
            write_scans_tsv,
        )

        scans_dict = create_scans_dict(
            clinical_data_dir=clinical_data_dir,
            study_name=StudyName.OASIS,
            clinical_specifications_folder=Path(__file__).parents[1] / "specifications",
            bids_ids=bids_ids,
            name_column_ids="ID",
            name_column_ses="",
            ses_dict=sessions,
        )
        write_scans_tsv(bids_dir, bids_ids, scans_dict)

    def _create_modality_agnostic_files(self, bids_dir: Path):
        from clinica.iotools.bids_utils import StudyName, write_modality_agnostic_files

        readme_data = {
            "link": "https://sites.wustl.edu/oasisbrains/#access",
            "desc": (
                "This set consists of a cross-sectional collection of 416 subjects aged 18 to 96. For each subject, 3 "
                "or 4 individual T1-weighted MRI scans obtained in single scan sessions are included. The subjects are "
                "all right-handed and include both men and women. 100 of the included subjects over the age of 60 have "
                "been clinically diagnosed with very mild to moderate Alzheimer’s disease (AD). Additionally, a "
                "reliability data set is included containing 20 nondemented subjects imaged on a subsequent visit "
                "within 90 days of their initial session."
            ),
        }
        write_modality_agnostic_files(
            study_name=StudyName.OASIS,
            readme_data=readme_data,
            bids_dir=bids_dir,
        )

    @staticmethod
    def convert_single_subject(subj_folder: Path, dest_dir: Path):
        from clinica.cmdline import setup_clinica_logging
        from clinica.iotools.bids_utils import StudyName, bids_id_factory
        from clinica.utils.stream import cprint

        # This function is executed in a multiprocessing context
        # such that we need to re-configure the clinica logger in the child processes.
        # Note that logging messages could easily be lost (for example when logging
        # to a file from two different processes). A better solution would be to
        # implement a logging process consuming logging messages from a multiprocessing.Queue...
        setup_clinica_logging("INFO")

        t1_folder = subj_folder / "PROCESSED" / "MPRAGE" / "SUBJ_111"
        cprint(f"Converting {subj_folder.name}", lvl="info")
        participant_id = bids_id_factory(StudyName.OASIS).from_original_study_id(
            subj_folder.name
        )
        bids_subj_folder = dest_dir / participant_id
        if not bids_subj_folder.is_dir():
            bids_subj_folder.mkdir(parents=True)
        session_folder = bids_subj_folder / "ses-M000"
        if not session_folder.is_dir():
            (session_folder / "anat").mkdir(parents=True, exist_ok=True)
        # In order do convert the Analyze format to Nifti the path to the .img file is required
        try:
            img_file_path = next(t1_folder.glob("*.img"))
        except StopIteration:
            msg = f"No file ending in .img found in {t1_folder}."
            cprint(msg, lvl="error")
            raise FileNotFoundError(msg)
        nb.save(
            _get_image_with_good_orientation(img_file_path),
            session_folder / "anat" / f"{participant_id}_ses-M000_T1w.nii.gz",
        )

    def convert_images(
        self, source_dir: Path, dest_dir: Path, n_procs: Optional[int] = 1
    ):
        """Convert T1w images to BIDS.

        Parameters
        ----------
        source_dir: path to the OASIS dataset

        dest_dir: path to the BIDS directory

        n_procs : int, optional
            The requested number of processes.
            If specified, it should be between 1 and the number of available CPUs.
            Default=1.

        Notes
        -----
        Previous version of this method used mri_convert from FreeSurfer to convert
        Analyze data from OASIS-1. To remove this strong dependency, NiBabel is used instead.
        """
        from functools import partial
        from multiprocessing import Pool

        if not dest_dir.exists():
            dest_dir.mkdir(parents=True)

        subjects_folders = [
            path
            for path in source_dir.rglob("OAS1_*")
            if path.is_dir() and path.name.endswith("_MR1")
        ]
        func = partial(self.convert_single_subject, dest_dir=dest_dir)
        # If n_procs==1 do not rely on a Process Pool to enable classical debugging
        if n_procs == 1:
            for folder in subjects_folders:
                func(folder)
        else:
            with Pool(processes=n_procs) as pool:
                pool.map(func, subjects_folders)


def _get_image_with_good_orientation(image_path: Path) -> nb.Nifti1Image:
    # First, convert to Nifti so that we can extract the s_form with NiBabel
    # (NiBabel creates an 'Spm2AnalyzeImage' object that does not contain 'get_sform' method
    img_with_wrong_orientation_analyze = nb.load(image_path)

    # OASIS-1 images have the same header but sform is incorrect
    # To solve this issue, we use header from images converted with FreeSurfer
    # to generate a 'clean hard-coded' header
    # affine:
    # [[   0.    0.   -1.   80.]
    #  [   1.    0.    0. -128.]
    #  [   0.    1.    0. -128.]
    #  [   0.    0.    0.    1.]]
    # fmt: off
    affine = np.array(
        [
            0, 0, -1, 80,
            1, 0, 0, -128,
            0, 1, 0, -128,
            0, 0, 0, 1
        ]
    ).reshape(4, 4)
    # fmt: on
    s_form = affine.astype(np.int16)

    hdr = nb.Nifti1Header()
    hdr.set_data_shape((256, 256, 160))
    hdr.set_data_dtype(np.int16)
    hdr["bitpix"] = 16
    hdr.set_sform(s_form, code="scanner")
    hdr.set_qform(s_form, code="scanner")
    hdr["extents"] = 16384
    hdr["xyzt_units"] = 10

    img_with_good_orientation_nifti = nb.Nifti1Image(
        np.round(img_with_wrong_orientation_analyze.get_fdata(dtype="float32")).astype(
            np.int16
        ),
        s_form,
        header=hdr,
    )
    # Header correction to obtain dim0 = 3
    return nb.funcs.four_to_three(img_with_good_orientation_nifti)[0]



"""This module contains utilities used by the DWIConnectome pipeline."""

from pathlib import Path
from typing import List, Optional

__all__ = [
    "get_luts",
    "get_conversion_luts",
    "get_containers",
    "get_caps_filenames",
    "print_begin_pipeline",
    "print_end_pipeline",
    "convert_flirt_to_mrtrix_transformation",
]


def get_luts() -> List[str]:
    from clinica.utils.check_dependency import get_freesurfer_home

    return [
        str(get_freesurfer_home() / "FreeSurferColorLUT.txt"),
        str(get_freesurfer_home() / "FreeSurferColorLUT.txt"),
    ]


def get_conversion_luts() -> List[str]:
    from pathlib import Path

    path_to_mappings = (
        Path(__file__).resolve().parent.parent.parent.parent / "resources" / "mappings"
    )
    resulting_paths = []
    for filename in ("fs_default.txt", "fs_a2009s.txt"):
        file_path = path_to_mappings / filename
        if not file_path.is_file():
            file_path = _download_mrtrix3_file(filename, path_to_mappings)
        resulting_paths.append(str(file_path))
    return resulting_paths


def _download_mrtrix3_file(filename: str, path_to_mappings: Path) -> Path:
    from clinica.utils.inputs import RemoteFileStructure, fetch_file
    from clinica.utils.stream import cprint

    try:
        return fetch_file(
            RemoteFileStructure(
                filename=filename,
                url="https://raw.githubusercontent.com/MRtrix3/mrtrix3/master/share/mrtrix3/labelconvert/",
                checksum=_get_checksum_for_filename(filename),
            ),
            path_to_mappings,
        )
    except IOError as err:
        error_msg = f"Unable to download required MRTRIX mapping ({filename}) for processing: {err}"
        cprint(msg=error_msg, lvl="error")
        raise IOError(error_msg)


def _get_checksum_for_filename(filename: str) -> str:
    if filename == "fs_default.txt":
        return "a8d561694887a1ca8d9df223aa5ef861b6c79d43ce9ed93835b9ce8aadc331b1"
    if filename == "fs_a2009s.txt":
        return "40b0d4d77bde7e1d265439347af5b30cc973748c1a88d203d7044cb35b3863e1"
    raise ValueError(f"File name {filename} is not supported.")


def get_containers(subjects: List[str], sessions: List[str]) -> List[str]:
    from pathlib import Path

    return [
        str(Path("subjects") / subject / session / "dwi")
        for subject, session in zip(subjects, sessions)
    ]


def get_caps_filenames(dwi_file: str) -> tuple:
    import re

    error_msg = f"Input filename {dwi_file} is not in a CAPS compliant format."
    if (
        m := re.search(
            r"/(sub-[a-zA-Z0-9]+_ses-[a-zA-Z0-9]+.*_desc-preproc*)_dwi", dwi_file
        )
    ) is None:
        raise ValueError(error_msg)
    source_file_caps = m.group(1)
    if (
        m := re.search(
            r"/(sub-[a-zA-Z0-9]+_ses-[a-zA-Z0-9]+.*)_space-[a-zA-Z0-9]+_desc-preproc_dwi",
            dwi_file,
        )
    ) is None:
        raise ValueError(error_msg)
    source_file_bids = m.group(1)

    response = f"{source_file_caps}_model-CSD_responseFunction.txt"
    fod = f"{source_file_caps}_model-CSD_diffmodel.nii.gz"
    tracts = f"{source_file_caps}_model-CSD_tractography.tck"
    nodes = [
        f"{source_file_caps}_atlas-{atlas}_parcellation.nii.gz"
        for atlas in ("desikan", "destrieux")
    ]
    connectomes = [
        f"{source_file_bids}_model-CSD_atlas-{atlas}_connectivity.tsv"
        for atlas in ("desikan", "destrieux")
    ]

    return response, fod, tracts, nodes, connectomes


def print_begin_pipeline(in_bids_or_caps_file: str) -> None:
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    print_begin_image(get_subject_id(in_bids_or_caps_file))


def print_end_pipeline(in_bids_or_caps_file: str, final_file: str) -> None:
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_end_image

    print_end_image(get_subject_id(in_bids_or_caps_file))


def _check_file_presence(
    file: Path,
) -> None:
    if not file.is_file():
        raise FileNotFoundError(f"The file {file} was not found.")


def _get_transformation_cmd(
    source_image: Path,
    reference_image: Path,
    flirt_matrix: Path,
    mrtrix_matrix: Path,
) -> str:
    for file in (source_image, reference_image, flirt_matrix):
        _check_file_presence(file)

    return f"transformconvert {flirt_matrix} {source_image} {reference_image} flirt_import {mrtrix_matrix}"


def convert_flirt_to_mrtrix_transformation(
    source_image: Path,
    reference_image: Path,
    flirt_matrix: Path,
    name_output_matrix: Optional[str] = None,
) -> Path:
    """Convert flirt matrix to mrtrix matrix.

    This function converts a transformation matrix produced by FSL's flirt
    command into a format usable by MRtrix. The output of this function
    is usually for the mrtransform command.

    Parameters
    ----------
    source_image : Path
        File containing the source image used in FSL flirt with the -in flag.

    reference_image : Path
        File containing the reference image used in FSL flirt with the -ref flag.

    flirt_matrix : Path
        File containing the transformation matrix obtained by FSL flirt.

    name_output_matrix : Path, optional
        Name of the output matrix. Defaults to "mrtrix_matrix.mat".

    Returns
    -------
    mrtrix_matrix : Path
        Transformation matrix in MRtrix format.
    """
    import os

    from clinica.utils.check_dependency import ThirdPartySoftware, check_software

    check_software(ThirdPartySoftware.MRTRIX)

    name_output_matrix = name_output_matrix or "mrtrix_matrix.mat"
    mrtrix_matrix = Path(name_output_matrix)
    mrtrix_matrix = mrtrix_matrix.resolve()

    cmd = _get_transformation_cmd(
        source_image, reference_image, flirt_matrix, mrtrix_matrix
    )
    os.system(cmd)

    return mrtrix_matrix



from __future__ import annotations

from os import PathLike
from pathlib import Path
from typing import Optional, Sequence

from pydra.engine import Workflow
from pydra.mark import annotate, task

__all__ = ["read_bids", "read_bids_dataset", "read_bids_files"]


@task
@annotate(
    {
        "return": {
            "dataset_description": Optional[dict],
            "participant_ids": list[str],
            "session_ids": Optional[list[str]],
        }
    }
)
def read_bids_dataset(dataset_path: PathLike):
    dataset_path = Path(dataset_path).resolve()
    description_file = dataset_path / "dataset_description.json"
    dataset_description = (
        json.load(description_file.read_text()) if description_file.exists() else None
    )

    try:
        _ = next(dataset_path.glob("*/ses-*"))
        multi_sessions = True
    except StopIteration:
        multi_sessions = False

    if multi_sessions:
        visits = dataset_path.glob("sub-*/ses-*")
        participant_ids, session_ids = list(
            map(
                list,
                zip(
                    *(
                        str(visit.relative_to(dataset_path)).split("/")
                        for visit in visits
                    )
                ),
            )
        )
    else:
        visits = dataset_path.glob("sub-*")
        participant_ids = sorted(
            str(visit.relative_to(dataset_path)) for visit in visits
        )
        session_ids = None

    return dataset_description, participant_ids, session_ids


@task
@annotate({"return": {"files": list[Path]}})
def read_bids_files(
    dataset_path: PathLike,
    participant_ids: Sequence[str] | None = None,
    session_ids: Sequence[str] | None = None,
    datatype: str | None = None,
    suffix: str | None = None,
    extension: str | None = None,
):
    dataset_path = Path(dataset_path).resolve()
    datatype = datatype or "*"
    suffix = suffix or "*"
    extension = extension or "*"
    files = []

    if all([participant_ids, session_ids]):
        for participant_id, session_id in zip([participant_ids, session_ids]):
            dir_pattern = f"{participant_id}/{session_id}/{datatype}"
            name_pattern = f"{participant_id}_{session_id}*_{suffix}.{extension}"
            file_pattern = f"{dir_pattern}/{name_pattern}"
            files += sorted(dataset_path.glob(file_pattern))
    elif participant_ids:
        for participant_id in participant_ids:
            dir_pattern = f"{participant_id}/**/{datatype}"
            name_pattern = f"{participant_id}*_{suffix}.{extension}"
            file_pattern = f"{dir_pattern}/{name_pattern}"
            files += sorted(dataset_path.glob(file_pattern))
    else:
        dir_pattern = f"**/{datatype}"
        name_pattern = f"*_{suffix}.{extension}"
        file_pattern = f"{dir_pattern}/{name_pattern}"
        files += sorted(dataset_path.glob(file_pattern))

    return files


def read_bids(output_queries: dict, **kwargs) -> Workflow:
    workflow = Workflow(name="read_bids", input_spec=["dataset_path"], **kwargs)

    workflow.add(
        read_bids_dataset(
            name="read_bids_dataset", dataset_path=workflow.lzin.dataset_path
        )
    )
    connections = {
        "dataset_description": workflow.read_bids_dataset.lzout.dataset_description,
        "participant_ids": workflow.read_bids_dataset.lzout.participant_ids,
        "session_ids": workflow.read_bids_dataset.lzout.session_ids,
    }

    for output_name, bids_query in output_queries.items():
        task_ = read_bids_files(
            name=f"read_{output_name}",
            dataset_path=workflow.lzin.dataset_path,
            participant_ids=workflow.read_bids_dataset.lzout.participant_ids,
            session_ids=workflow.read_bids_dataset.lzout.session_ids,
            **bids_query,
        )
        workflow.add(task_)
        connections.update({output_name: task_.lzout.files})

    workflow.set_output(connections=connections)

    return workflow



from typing import List

from nipype import config

from clinica.pipelines.pet.engine import PETPipeline

# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class PETVolume(PETPipeline):
    """PETVolume - Volume-based processing of PET images using SPM.

    Returns:
        A clinica pipeline object containing the PETVolume pipeline.
    """

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        from clinica.utils.atlas import T1AndPetVolumeAtlasName
        from clinica.utils.group import check_group_label

        super()._check_pipeline_parameters()
        self.parameters.setdefault("group_label", None)
        check_group_label(self.parameters["group_label"])
        self.parameters.setdefault("pvc_psf_tsv", None)
        self.parameters.setdefault("mask_tissues", [1, 2, 3])
        self.parameters.setdefault("mask_threshold", 0.3)
        self.parameters.setdefault("pvc_mask_tissues", [1, 2, 3])
        self.parameters.setdefault("smooth", [8])
        self.parameters.setdefault("atlases", T1AndPetVolumeAtlasName)

    def _check_custom_dependencies(self) -> None:
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) input fields name.
        """
        return [
            "pet_image",
            "t1_image_native",
            "mask_tissues",
            "pvc_mask_tissues",
            "psf",
            "flow_fields",
            "dartel_template",
            "reference_mask",
        ]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) output fields name.
        """
        return [
            "pet_t1_native",
            "pet_mni",
            "pet_suvr",
            "binary_mask",
            "pet_suvr_masked",
            "pet_suvr_masked_smoothed",
            "pet_pvc",
            "pet_pvc_mni",
            "pet_pvc_suvr",
            "pet_pvc_suvr_masked",
            "pet_pvc_suvr_masked_smoothed",
            "atlas_statistics",
            "pvc_atlas_statistics",
        ]

    def _build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.iotools.utils.data_handling import (
            check_relative_volume_location_in_world_coordinate_system,
        )
        from clinica.pipelines.pet.utils import get_suvr_mask, read_psf_information
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.filemanip import save_participants_sessions
        from clinica.utils.input_files import (
            T1W_NII,
            t1_volume_deformation_to_template,
            t1_volume_final_group_template,
            t1_volume_native_tpm,
            t1_volume_native_tpm_in_mni,
        )
        from clinica.utils.inputs import clinica_file_reader, clinica_group_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import (
            print_groups_in_caps_directory,
            print_images_to_process,
        )

        # Check that group already exists
        if not (
            self.caps_directory / "groups" / f"group-{self.parameters['group_label']}"
        ).exists():
            print_groups_in_caps_directory(self.caps_directory)
            raise ClinicaException(
                f"Group {self.parameters['group_label']} does not exist. "
                "Did you run t1-volume or t1-volume-create-dartel pipeline?"
            )

        # Tissues DataGrabber
        # ====================
        all_errors = []

        # Grab reference mask
        reference_mask_file = str(
            get_suvr_mask(self.parameters["suvr_reference_region"])
        )

        # PET from BIDS directory
        try:
            pet_bids, _ = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.bids_directory,
                self._get_pet_scans_query(),
            )
        except ClinicaException as e:
            all_errors.append(e)

        # Native T1w-MRI
        try:
            t1w_bids, _ = clinica_file_reader(
                self.subjects, self.sessions, self.bids_directory, T1W_NII
            )
        except ClinicaException as e:
            all_errors.append(e)

        # mask_tissues
        tissues_input = []
        for tissue_number in self.parameters["mask_tissues"]:
            try:
                current_file, _ = clinica_file_reader(
                    self.subjects,
                    self.sessions,
                    self.caps_directory,
                    t1_volume_native_tpm_in_mni(tissue_number, False),
                )
                tissues_input.append(current_file)
            except ClinicaException as e:
                all_errors.append(e)
        # Tissues_input has a length of len(self.parameters['mask_tissues']). Each of these elements has a size of
        # len(self.subjects). We want the opposite: a list of size len(self.subjects) whose elements have a size of
        # len(self.parameters['mask_tissues']. The trick is to iter on elements with zip(*my_list)
        tissues_input_final = []
        for subject_tissue_list in zip(*tissues_input):
            tissues_input_final.append(subject_tissue_list)
        tissues_input = tissues_input_final

        # Flowfields
        try:
            flowfields_caps, _ = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                t1_volume_deformation_to_template(self.parameters["group_label"]),
            )
        except ClinicaException as e:
            all_errors.append(e)

        # Dartel Template
        try:
            final_template = clinica_group_reader(
                self.caps_directory,
                t1_volume_final_group_template(self.parameters["group_label"]),
            )
        except ClinicaException as e:
            all_errors.append(e)

        if self.parameters["pvc_psf_tsv"] is not None:
            iterables_psf = read_psf_information(
                self.parameters["pvc_psf_tsv"],
                self.subjects,
                self.sessions,
                self.parameters["acq_label"],
            )
            self.parameters["apply_pvc"] = True
        else:
            iterables_psf = [[]] * len(self.subjects)
            self.parameters["apply_pvc"] = False

        if self.parameters["apply_pvc"]:
            # pvc tissues input
            pvc_tissues_input = []
            for tissue_number in self.parameters["pvc_mask_tissues"]:
                try:
                    current_file, _ = clinica_file_reader(
                        self.subjects,
                        self.sessions,
                        self.caps_directory,
                        t1_volume_native_tpm(tissue_number),
                    )
                    pvc_tissues_input.append(current_file)
                except ClinicaException as e:
                    all_errors.append(e)

            if len(all_errors) == 0:
                pvc_tissues_input_final = []
                for subject_tissue_list in zip(*pvc_tissues_input):
                    pvc_tissues_input_final.append(subject_tissue_list)
                pvc_tissues_input = pvc_tissues_input_final
        else:
            pvc_tissues_input = []

        if len(all_errors) > 0:
            error_message = "Clinica faced error(s) while trying to read files in your CAPS/BIDS directories.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaException(error_message)

        check_relative_volume_location_in_world_coordinate_system(
            "T1w-MRI",
            t1w_bids,
            self.parameters["acq_label"] + " PET",
            pet_bids,
            self.bids_directory,
            self.parameters["acq_label"],
            skip_question=self.parameters["skip_question"],
        )

        # Save subjects to process in <WD>/<Pipeline.name>/participants.tsv
        save_participants_sessions(
            self.subjects, self.sessions, self.base_dir / self.name
        )

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint(
                f"List available in {self.base_dir / self.name / 'participants.tsv'}"
            )
            cprint("The pipeline will last approximately 10 minutes per image.")

        read_input_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
            iterables=[
                ("pet_image", pet_bids),
                ("t1_image_native", t1w_bids),
                ("mask_tissues", tissues_input),
                ("psf", iterables_psf),
                ("flow_fields", flowfields_caps),
                ("pvc_mask_tissues", pvc_tissues_input),
            ],
            synchronize=True,
        )

        read_input_node.inputs.reference_mask = reference_mask_file
        read_input_node.inputs.dartel_template = final_template

        self.connect(
            [
                (
                    read_input_node,
                    self.input_node,
                    [
                        ("pet_image", "pet_image"),
                        ("t1_image_native", "t1_image_native"),
                        ("mask_tissues", "mask_tissues"),
                        ("flow_fields", "flow_fields"),
                        ("dartel_template", "dartel_template"),
                        ("reference_mask", "reference_mask"),
                        ("psf", "psf"),
                        ("pvc_mask_tissues", "pvc_mask_tissues"),
                    ],
                )
            ]
        )

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import re

        import nipype.interfaces.io as nio
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import zip_nii
        from clinica.utils.nipype import container_from_filename, fix_join

        # Find container path from pet filename
        # =====================================
        container_path = npe.Node(
            nutil.Function(
                input_names=["bids_or_caps_filename"],
                output_names=["container"],
                function=container_from_filename,
            ),
            name="container_path",
        )
        container_path.inputs.threshold = self.parameters["mask_threshold"]

        # Writing all images into CAPS
        # ============================
        write_images_node = npe.Node(name="write_caps_node", interface=nio.DataSink())
        write_images_node.inputs.base_directory = str(self.caps_directory)
        write_images_node.inputs.parameterization = False
        write_images_node.inputs.regexp_substitutions = [
            (r"(.*/)pet_t1_native/r(sub-.*)(\.nii(\.gz)?)$", r"\1\2_space-T1w_pet\3"),
            (
                r"(.*/)pet_pvc/pvc-rbv_r(sub-.*)(\.nii(\.gz)?)$",
                r"\1\2_space-T1w_pvc-rbv_pet\3",
            ),
            (
                r"(.*/)pet_mni/wr(sub-.*)(\.nii(\.gz)?)$",
                r"\1\2_space-Ixi549Space_pet\3",
            ),
            (
                r"(.*/)pet_pvc_mni/wpvc-rbv_r(sub-.*)(\.nii(\.gz)?)$",
                r"\1\2_space-Ixi549Space_pvc-rbv_pet\3",
            ),
            (
                r"(.*/)pet_suvr/suvr_wr(sub-.*)(\.nii(\.gz)?)$",
                r"\1\2_space-Ixi549Space_suvr-"
                + re.escape(self.parameters["suvr_reference_region"])
                + r"_pet\3",
            ),
            (
                r"(.*/)pet_pvc_suvr/suvr_wpvc-rbv_r(sub-.*)(\.nii(\.gz)?)$",
                r"\1\2_space-Ixi549Space_pvc-rbv_suvr-"
                + re.escape(self.parameters["suvr_reference_region"])
                + r"_pet\3",
            ),
            (
                r"(.*/)pet_suvr_masked/masked_suvr_wr(sub-.*)(\.nii(\.gz)?)$",
                r"\1\2_space-Ixi549Space_suvr-"
                + re.escape(self.parameters["suvr_reference_region"])
                + r"_mask-brain_pet\3",
            ),
            (
                r"(.*/)pet_pvc_suvr_masked/masked_suvr_wpvc-rbv_r(sub-.*)(\.nii(\.gz)?)$",
                r"\1\2_space-Ixi549Space_pvc-rbv_suvr-"
                + re.escape(self.parameters["suvr_reference_region"])
                + r"_mask-brain_pet\3",
            ),
            (
                r"(.*/)pet_suvr_masked_smoothed/(fwhm-[0-9]+mm)_masked_suvr_wr(sub-.*)(\.nii(\.gz)?)$",
                r"\1\3_space-Ixi549Space_suvr-"
                + re.escape(self.parameters["suvr_reference_region"])
                + r"_mask-brain_\2_pet\4",
            ),
            (
                r"(.*/)pet_pvc_suvr_masked_smoothed/(fwhm-[0-9]+mm)_masked_suvr_wpvc-rbv_r(sub-.*)(\.nii(\.gz)?)$",
                r"\1\3_space-Ixi549Space_pvc-rbv_suvr-"
                + re.escape(self.parameters["suvr_reference_region"])
                + r"_mask-brain_\2_pet\4",
            ),
            (
                r"(.*/)binary_mask/(sub-.*_T1w_).*(space-[a-zA-Z0-9]+).*(_brainmask\.nii(\.gz)?)$",
                r"\1\2\3\4",
            ),
        ]

        # Writing atlas statistics into CAPS
        # ==================================
        write_atlas_node = npe.Node(name="write_atlas_node", interface=nio.DataSink())
        write_atlas_node.inputs.base_directory = str(self.caps_directory)
        write_atlas_node.inputs.parameterization = False
        write_atlas_node.inputs.regexp_substitutions = [
            (
                r"(.*/atlas_statistics/)suvr_wr(sub-.*)(_statistics\.tsv)$",
                r"\1\2"
                + r"_suvr-"
                + re.escape(self.parameters["suvr_reference_region"])
                + r"\3",
            ),
            (
                r"(.*/)pvc_(atlas_statistics/)suvr_wpvc-rbv_r(sub-.*)(_statistics\.tsv)$",
                r"\1\2\3"
                + r"_pvc-rbv_suvr-"
                + re.escape(self.parameters["suvr_reference_region"])
                + r"\4",
            ),
        ]
        self.connect(
            [
                (
                    self.input_node,
                    container_path,
                    [("pet_image", "bids_or_caps_filename")],
                ),
                (
                    container_path,
                    write_images_node,
                    [
                        (
                            (
                                "container",
                                fix_join,
                                "pet",
                                "preprocessing",
                                f"group-{self.parameters['group_label']}",
                            ),
                            "container",
                        )
                    ],
                ),
                (
                    self.output_node,
                    write_images_node,
                    [
                        (("pet_t1_native", zip_nii, True), "pet_t1_native"),
                        (("pet_mni", zip_nii, True), "pet_mni"),
                        (("pet_suvr", zip_nii, True), "pet_suvr"),
                        (("binary_mask", zip_nii, True), "binary_mask"),
                        (("pet_suvr_masked", zip_nii, True), "pet_suvr_masked"),
                        (
                            ("pet_suvr_masked_smoothed", zip_nii, True),
                            "pet_suvr_masked_smoothed",
                        ),
                        (("pet_pvc", zip_nii, True), "pet_pvc"),
                        (("pet_pvc_mni", zip_nii, True), "pet_pvc_mni"),
                        (("pet_pvc_suvr", zip_nii, True), "pet_pvc_suvr"),
                        (("pet_pvc_suvr_masked", zip_nii, True), "pet_pvc_suvr_masked"),
                        (
                            ("pet_pvc_suvr_masked_smoothed", zip_nii, True),
                            "pet_pvc_suvr_masked_smoothed",
                        ),
                    ],
                ),
                (
                    container_path,
                    write_atlas_node,
                    [
                        (
                            (
                                "container",
                                fix_join,
                                "pet",
                                "preprocessing",
                                f"group-{self.parameters['group_label']}",
                            ),
                            "container",
                        )
                    ],
                ),
                (
                    self.output_node,
                    write_atlas_node,
                    [
                        ("atlas_statistics", "atlas_statistics"),
                        ("pvc_atlas_statistics", "pvc_atlas_statistics"),
                    ],
                ),
            ]
        )

    def _build_core_nodes(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.spm as spm
        import nipype.interfaces.spm.utils as spmutils
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces.petpvc import PETPVC

        from clinica.utils.filemanip import unzip_nii
        from clinica.utils.spm import use_spm_standalone_if_available

        from .tasks import (
            apply_binary_mask_task,
            compute_atlas_statistics_task,
            create_binary_mask_task,
            create_pvc_mask_task,
            normalize_to_reference_task,
        )
        from .utils import (
            build_pet_pvc_name,
            get_from_list,
            init_input_node,
        )

        use_spm_standalone_if_available()

        # Initialize pipeline
        # ===================
        init_node = npe.Node(
            interface=nutil.Function(
                input_names=["pet_nii"],
                output_names=["pet_nii"],
                function=init_input_node,
            ),
            name="init_pipeline",
        )

        # Unzipping
        # =========
        unzip_pet_image = npe.Node(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_pet_image",
        )

        unzip_t1_image_native = npe.Node(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_t1_image_native",
        )

        unzip_flow_fields = npe.Node(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_flow_fields",
        )

        unzip_dartel_template = npe.Node(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_dartel_template",
        )

        unzip_reference_mask = npe.Node(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_reference_mask",
        )

        unzip_mask_tissues = npe.MapNode(
            nutil.Function(
                input_names=["in_file"], output_names=["out_file"], function=unzip_nii
            ),
            name="unzip_mask_tissues",
            iterfield=["in_file"],
        )

        # Coregister PET into T1 native space
        # ===================================
        coreg_pet_t1 = npe.Node(spm.Coregister(), name="coreg_pet_t1")

        # Spatially normalize PET into MNI
        # ================================
        dartel_mni_reg = npe.Node(spm.DARTELNorm2MNI(), name="dartel_mni_reg")
        dartel_mni_reg.inputs.modulate = False
        dartel_mni_reg.inputs.fwhm = 0

        # Reslice reference region mask into PET
        # ======================================
        reslice = npe.Node(spmutils.Reslice(), name="reslice")

        # Normalize PET values according to reference region
        # ==================================================
        norm_to_ref = npe.Node(
            nutil.Function(
                input_names=["pet_image", "region_mask"],
                output_names=["suvr_pet_path"],
                function=normalize_to_reference_task,
            ),
            name="norm_to_ref",
        )

        # Create binary mask from segmented tissues
        # =========================================
        binary_mask = npe.Node(
            nutil.Function(
                input_names=["tissues", "threshold"],
                output_names=["out_mask"],
                function=create_binary_mask_task,
            ),
            name="binary_mask",
        )
        binary_mask.inputs.threshold = self.parameters["mask_threshold"]

        # Mask PET image
        # ==============
        apply_mask = npe.Node(
            nutil.Function(
                input_names=["image", "binary_mask"],
                output_names=["masked_image_path"],
                function=apply_binary_mask_task,
            ),
            name="apply_mask",
        )

        # Smoothing
        # =========
        if self.parameters["smooth"] is not None and len(self.parameters["smooth"]) > 0:
            smoothing_node = npe.MapNode(
                spm.Smooth(), name="smoothing_node", iterfield=["fwhm", "out_prefix"]
            )
            smoothing_node.inputs.fwhm = [[x, x, x] for x in self.parameters["smooth"]]
            smoothing_node.inputs.out_prefix = [
                "fwhm-" + str(x) + "mm_" for x in self.parameters["smooth"]
            ]
            self.connect(
                [
                    (apply_mask, smoothing_node, [("masked_image_path", "in_files")]),
                    (
                        smoothing_node,
                        self.output_node,
                        [("smoothed_files", "pet_suvr_masked_smoothed")],
                    ),
                ]
            )
        else:
            self.output_node.inputs.pet_suvr_masked_smoothed = [[]]

        # Atlas Statistics
        # ================
        atlas_stats_node = npe.MapNode(
            nutil.Function(
                input_names=["image", "atlas_names"],
                output_names=["atlas_statistics"],
                function=compute_atlas_statistics_task,
            ),
            name="atlas_stats_node",
            iterfield=["in_image"],
        )
        atlas_stats_node.inputs.atlas_names = self.parameters["atlases"]

        self.connect(
            [
                (self.input_node, init_node, [("pet_image", "pet_nii")]),
                (init_node, unzip_pet_image, [("pet_nii", "in_file")]),
                (
                    self.input_node,
                    unzip_t1_image_native,
                    [("t1_image_native", "in_file")],
                ),
                (self.input_node, unzip_flow_fields, [("flow_fields", "in_file")]),
                (
                    self.input_node,
                    unzip_dartel_template,
                    [("dartel_template", "in_file")],
                ),
                (
                    self.input_node,
                    unzip_reference_mask,
                    [("reference_mask", "in_file")],
                ),
                (self.input_node, unzip_mask_tissues, [("mask_tissues", "in_file")]),
                (unzip_pet_image, coreg_pet_t1, [("out_file", "source")]),
                (unzip_t1_image_native, coreg_pet_t1, [("out_file", "target")]),
                (unzip_flow_fields, dartel_mni_reg, [("out_file", "flowfield_files")]),
                (
                    unzip_dartel_template,
                    dartel_mni_reg,
                    [("out_file", "template_file")],
                ),
                (unzip_reference_mask, reslice, [("out_file", "in_file")]),
                (unzip_mask_tissues, binary_mask, [("out_file", "tissues")]),
                (
                    coreg_pet_t1,
                    dartel_mni_reg,
                    [("coregistered_source", "apply_to_files")],
                ),
                (dartel_mni_reg, reslice, [("normalized_files", "space_defining")]),
                (dartel_mni_reg, norm_to_ref, [("normalized_files", "pet_image")]),
                (reslice, norm_to_ref, [("out_file", "region_mask")]),
                (norm_to_ref, apply_mask, [("suvr_pet_path", "image")]),
                (binary_mask, apply_mask, [("out_mask", "binary_mask")]),
                (norm_to_ref, atlas_stats_node, [("suvr_pet_path", "image")]),
                (
                    coreg_pet_t1,
                    self.output_node,
                    [("coregistered_source", "pet_t1_native")],
                ),
                (dartel_mni_reg, self.output_node, [("normalized_files", "pet_mni")]),
                (norm_to_ref, self.output_node, [("suvr_pet_path", "pet_suvr")]),
                (binary_mask, self.output_node, [("out_mask", "binary_mask")]),
                (
                    apply_mask,
                    self.output_node,
                    [("masked_image_path", "pet_suvr_masked")],
                ),
                (
                    atlas_stats_node,
                    self.output_node,
                    [("atlas_statistics", "atlas_statistics")],
                ),
            ]
        )

        # PVC
        # ==========
        if self.parameters["apply_pvc"]:
            # Unzipping
            # =========
            unzip_pvc_mask_tissues = npe.MapNode(
                nutil.Function(
                    input_names=["in_file"],
                    output_names=["out_file"],
                    function=unzip_nii,
                ),
                name="unzip_pvc_mask_tissues",
                iterfield=["in_file"],
            )

            # Creating Mask to use in PVC
            # ===========================
            pvc_mask = npe.Node(
                nutil.Function(
                    input_names=["tissues"],
                    output_names=["out_mask"],
                    function=create_pvc_mask_task,
                ),
                name="pvc_mask",
            )
            # PET PVC
            # =======
            petpvc = npe.Node(PETPVC(), name="pvc")
            petpvc.inputs.pvc = "RBV"
            petpvc.inputs.out_file = "pvc.nii"

            # Spatially normalize PET into MNI
            # ================================
            dartel_mni_reg_pvc = npe.Node(
                spm.DARTELNorm2MNI(), name="dartel_mni_reg_pvc"
            )
            dartel_mni_reg_pvc.inputs.modulate = False
            dartel_mni_reg_pvc.inputs.fwhm = 0

            # Reslice reference region mask into PET
            # ======================================
            reslice_pvc = npe.Node(spmutils.Reslice(), name="reslice_pvc")

            # Normalize PET values according to reference region
            # ==================================================
            norm_to_ref_pvc = npe.Node(
                nutil.Function(
                    input_names=["pet_image", "region_mask"],
                    output_names=["suvr_pet_path"],
                    function=normalize_to_reference_task,
                ),
                name="norm_to_ref_pvc",
            )

            # Mask PET image
            # ==============
            apply_mask_pvc = npe.Node(
                nutil.Function(
                    input_names=["image", "binary_mask"],
                    output_names=["masked_image_path"],
                    function=apply_binary_mask_task,
                ),
                name="apply_mask_pvc",
            )
            # Smoothing
            # =========
            if (
                self.parameters["smooth"] is not None
                and len(self.parameters["smooth"]) > 0
            ):
                smoothing_pvc = npe.MapNode(
                    spm.Smooth(), name="smoothing_pvc", iterfield=["fwhm", "out_prefix"]
                )
                smoothing_pvc.inputs.fwhm = [
                    [x, x, x] for x in self.parameters["smooth"]
                ]
                smoothing_pvc.inputs.out_prefix = [
                    "fwhm-" + str(x) + "mm_" for x in self.parameters["smooth"]
                ]
                self.connect(
                    [
                        (
                            apply_mask_pvc,
                            smoothing_pvc,
                            [("masked_image_path", "in_files")],
                        ),
                        (
                            smoothing_pvc,
                            self.output_node,
                            [("smoothed_files", "pet_pvc_suvr_masked_smoothed")],
                        ),
                    ]
                )
            else:
                self.output_node.inputs.pet_pvc_suvr_masked_smoothed = [[]]
            # Atlas Statistics
            # ================
            atlas_stats_pvc = npe.MapNode(
                nutil.Function(
                    input_names=["image", "atlas_names"],
                    output_names=["atlas_statistics"],
                    function=compute_atlas_statistics_task,
                ),
                name="atlas_stats_pvc",
                iterfield=["in_image"],
            )
            atlas_stats_pvc.inputs.in_atlas_names = self.parameters["atlases"]

            # Connection
            # ==========
            self.connect(
                [
                    (
                        self.input_node,
                        unzip_pvc_mask_tissues,
                        [("pvc_mask_tissues", "in_file")],
                    ),
                    (unzip_pvc_mask_tissues, pvc_mask, [("out_file", "tissues")]),
                    (
                        unzip_flow_fields,
                        dartel_mni_reg_pvc,
                        [("out_file", "flowfield_files")],
                    ),
                    (
                        unzip_dartel_template,
                        dartel_mni_reg_pvc,
                        [("out_file", "template_file")],
                    ),
                    (unzip_reference_mask, reslice_pvc, [("out_file", "in_file")]),
                    (
                        coreg_pet_t1,
                        petpvc,
                        [
                            ("coregistered_source", "in_file"),
                            (
                                ("coregistered_source", build_pet_pvc_name, "RBV"),
                                "out_file",
                            ),
                        ],
                    ),
                    (pvc_mask, petpvc, [("out_mask", "mask_file")]),
                    (
                        self.input_node,
                        petpvc,
                        [
                            (("psf", get_from_list, 0), "fwhm_x"),
                            (("psf", get_from_list, 1), "fwhm_y"),
                            (("psf", get_from_list, 2), "fwhm_z"),
                        ],
                    ),
                    (petpvc, dartel_mni_reg_pvc, [("out_file", "apply_to_files")]),
                    (
                        dartel_mni_reg_pvc,
                        reslice_pvc,
                        [("normalized_files", "space_defining")],
                    ),
                    (
                        dartel_mni_reg_pvc,
                        norm_to_ref_pvc,
                        [("normalized_files", "pet_image")],
                    ),
                    (reslice_pvc, norm_to_ref_pvc, [("out_file", "region_mask")]),
                    (norm_to_ref_pvc, apply_mask_pvc, [("suvr_pet_path", "image")]),
                    (binary_mask, apply_mask_pvc, [("out_mask", "binary_mask")]),
                    (norm_to_ref_pvc, atlas_stats_pvc, [("suvr_pet_path", "image")]),
                    (petpvc, self.output_node, [("out_file", "pet_pvc")]),
                    (
                        dartel_mni_reg_pvc,
                        self.output_node,
                        [("normalized_files", "pet_pvc_mni")],
                    ),
                    (
                        norm_to_ref_pvc,
                        self.output_node,
                        [("suvr_pet_path", "pet_pvc_suvr")],
                    ),
                    (
                        apply_mask_pvc,
                        self.output_node,
                        [("masked_image_path", "pet_pvc_suvr_masked")],
                    ),
                    (
                        atlas_stats_pvc,
                        self.output_node,
                        [("atlas_statistics", "pvc_atlas_statistics")],
                    ),
                ]
            )
        else:
            self.output_node.inputs.pet_pvc = [[]]
            self.output_node.inputs.pet_pvc_mni = [[]]
            self.output_node.inputs.pet_pvc_suvr = [[]]
            self.output_node.inputs.pet_pvc_suvr_masked = [[]]
            self.output_node.inputs.pvc_atlas_statistics = [[]]
            self.output_node.inputs.pet_pvc_suvr_masked_smoothed = [[]]



from pathlib import Path
from typing import List

from nipype import config

from clinica.pipelines.dwi.preprocessing.engine import DWIPreprocessingPipeline

cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class DwiPreprocessingUsingT1(DWIPreprocessingPipeline):
    """DWI Preprocessing using T1 image for susceptibility distortion step.

    Ideas for improvement:
        - Replace prepare_reference_b0 function by a first run of FSL eddy
        - Replace B0-T1w registration by FA-T1w registration
        - Replace os.system(cmd) by call of Nipype interface
        - Use promising sdcflows workflows and/or dMRIprep

    Warnings:
        - Do not use this pipeline if you have fieldmap data in your dataset.

    Note:
        Some reading regarding the reproducibility of FSL eddy command:
        https://www.jiscmail.ac.uk/cgi-bin/webadmin?A2=fsl;1ccf038f.1608

    Returns:
        A clinica pipeline object containing the DWIPreprocessingUsingT1 pipeline.
    """

    @staticmethod
    def get_processed_images(
        caps_directory: Path, subjects: List[str], sessions: List[str]
    ) -> List[str]:
        from clinica.utils.filemanip import extract_image_ids
        from clinica.utils.input_files import DWI_PREPROC_NII
        from clinica.utils.inputs import clinica_file_reader

        image_ids: List[str] = []
        if caps_directory.is_dir():
            preproc_files, _ = clinica_file_reader(
                subjects, sessions, caps_directory, DWI_PREPROC_NII, False
            )
            image_ids = extract_image_ids(preproc_files)
        return image_ids

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        super()._check_pipeline_parameters()
        self.parameters.setdefault("delete_cache", False)
        self.parameters.setdefault("random_seed", None)
        self.parameters.setdefault("double_precision", True)

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Returns
        -------
        List[str] :
            The list of inputs for the DwiPreprocessingUsingT1 pipeline namely:
                - "t1w_filename" : Path of the T1w image in BIDS format.
                - "dwi_filename": Path of the diffusion weighted image in BIDS format.
                - "dwi_json_filename" : Path of the DWI JSON file in BIDS format and containing
                  'TotalReadoutTime' and 'PhaseEncodingDirection' metadata (see BIDS specifications).
                - "b_vectors_filename" : Path of the b-vectors file in BIDS format.
                - "b_values_filename" : Path of the b-values file in BIDS format.
        """
        return [
            "t1w_filename",
            "dwi_filename",
            "dwi_json_filename",
            "b_vectors_filename",
            "b_values_filename",
        ]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Returns:
            List[str]: The list of outputs generated by the DwiPreprocessingUsingT1 pipeline namely:
                * preproc_dwi: Path of the preprocessed DWI
                * preproc_bvec: Path of the preprocessed bvec
                * preproc_bval: Path of the preprocessed bval
                * b0_mask: Path of the b0 brainmask
        """
        return ["preproc_dwi", "preproc_bvec", "preproc_bval", "b0_mask"]

    def filter_qc(self) -> tuple[list[str], list[str]]:
        from clinica.pipelines.dwi.preprocessing.utils import check_dwi_volume
        from clinica.pipelines.dwi.utils import DWIDataset
        from clinica.utils.bids import BIDSFileName
        from clinica.utils.input_files import (
            DWI_BVAL,
            DWI_BVEC,
            DWI_NII,
        )
        from clinica.utils.inputs import clinica_list_of_files_reader
        from clinica.utils.stream import cprint

        subjects = []
        sessions = []
        list_bids_files = clinica_list_of_files_reader(
            self.subjects,
            self.sessions,
            self.bids_directory,
            [DWI_NII, DWI_BVEC, DWI_BVAL],
            raise_exception=True,
        )
        for dwi_image_file, b_vectors_file, b_values_file in zip(*list_bids_files):
            dwi_image_file_bids = BIDSFileName.from_name(dwi_image_file)
            try:
                check_dwi_volume(
                    DWIDataset(
                        dwi=dwi_image_file,
                        b_values=b_values_file,
                        b_vectors=b_vectors_file,
                    )
                )
                subjects.append(f"sub-{dwi_image_file_bids.subject}")
                sessions.append(f"ses-{dwi_image_file_bids.session}")
            except (IOError, ValueError) as e:
                cprint(
                    f"Ignoring DWI scan for subject {dwi_image_file_bids.subject} "
                    f"and session {dwi_image_file_bids.session} for the following reason: {e}",
                    lvl="warning",
                )

        return subjects, sessions

    def _build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import save_participants_sessions
        from clinica.utils.input_files import (
            DWI_BVAL,
            DWI_BVEC,
            DWI_JSON,
            DWI_NII,
            T1W_NII,
        )
        from clinica.utils.inputs import clinica_list_of_files_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        list_bids_files = clinica_list_of_files_reader(
            self.subjects,
            self.sessions,
            self.bids_directory,
            [T1W_NII, DWI_NII, DWI_JSON, DWI_BVEC, DWI_BVAL],
            raise_exception=True,
        )

        save_participants_sessions(
            self.subjects, self.sessions, self.base_dir / self.name
        )

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint(
                f"List available in {self.base_dir / self.name / 'participants.tsv'}"
            )
            cprint(
                "Computational time will depend of the number of volumes in your DWI dataset and the use of CUDA."
            )

        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                (x, y) for x, y in zip(self.get_input_fields(), list_bids_files)
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        self.connect(
            [
                (read_node, self.input_node, [(x, x) for x in self.get_input_fields()]),
            ]
        )

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.io as nio
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.nipype import container_from_filename, fix_join

        from .tasks import rename_into_caps_task

        container_path = npe.Node(
            nutil.Function(
                input_names=["bids_or_caps_filename"],
                output_names=["container"],
                function=container_from_filename,
            ),
            name="container_path",
        )
        files_to_write_in_caps = [
            "dwi_preproc_filename",
            "b_values_preproc_filename",
            "b_vectors_preproc_filename",
            "b0_brain_mask_filename",
        ]

        rename_into_caps = npe.Node(
            nutil.Function(
                input_names=["dwi_filename"] + files_to_write_in_caps,
                output_names=[f"{x}_caps" for x in files_to_write_in_caps],
                function=rename_into_caps_task,
            ),
            name="rename_into_caps",
        )

        write_results = npe.Node(name="write_results", interface=nio.DataSink())
        write_results.inputs.base_directory = str(self.caps_directory)
        write_results.inputs.parameterization = False

        self.connect(
            [
                (
                    self.input_node,
                    container_path,
                    [("dwi_filename", "bids_or_caps_filename")],
                ),
                (self.input_node, rename_into_caps, [("dwi_filename", "dwi_filename")]),
                (
                    self.output_node,
                    rename_into_caps,
                    [
                        ("preproc_dwi", "dwi_preproc_filename"),
                        ("preproc_bval", "b_values_preproc_filename"),
                        ("preproc_bvec", "b_vectors_preproc_filename"),
                        ("b0_mask", "b0_brain_mask_filename"),
                    ],
                ),
                (
                    container_path,
                    write_results,
                    [(("container", fix_join, "dwi"), "container")],
                ),
                (
                    rename_into_caps,
                    write_results,
                    [
                        ("dwi_preproc_filename_caps", "preprocessing.@preproc_dwi"),
                        (
                            "b_values_preproc_filename_caps",
                            "preprocessing.@preproc_bval",
                        ),
                        (
                            "b_vectors_preproc_filename_caps",
                            "preprocessing.@preproc_bvec",
                        ),
                        ("b0_brain_mask_filename_caps", "preprocessing.@b0_mask"),
                    ],
                ),
            ]
        )

    def _build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.fsl as fsl
        import nipype.interfaces.mrtrix3 as mrtrix3
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.pipelines.dwi.preprocessing.tasks import compute_average_b0_task
        from clinica.pipelines.dwi.preprocessing.workflows import eddy_fsl_pipeline

        from .tasks import prepare_reference_b0_task
        from .utils import init_input_node, print_end_pipeline
        from .workflows import epi_pipeline

        init_node = npe.Node(
            interface=nutil.Function(
                input_names=self.get_input_fields(),
                output_names=[
                    "image_id",
                    "t1w_filename",
                    "dwi_filename",
                    "b_vectors_filename",
                    "b_values_filename",
                    "total_readout_time",
                    "phase_encoding_direction",
                ],
                function=init_input_node,
            ),
            name="0-InitNode",
        )

        # Prepare b0 image for further corrections
        prepare_b0 = npe.Node(
            name="PrepareB0",
            interface=nutil.Function(
                input_names=[
                    "dwi_filename",
                    "b_values_filename",
                    "b_vectors_filename",
                    "b_value_threshold",
                    "working_directory",
                ],
                output_names=[
                    "out_reference_b0",
                    "out_b0_dwi_merge",
                    "out_updated_bval",
                    "out_updated_bvec",
                ],
                function=prepare_reference_b0_task,
            ),
        )
        prepare_b0.inputs.b_value_threshold = self.parameters["low_bval"]
        prepare_b0.inputs.working_directory = str(self.base_dir)

        # Head-motion correction + Eddy-currents correction
        eddy_fsl = eddy_fsl_pipeline(
            base_dir=self.base_dir,
            use_cuda=self.parameters["use_cuda"],
            initrand=self.parameters["initrand"],
            compute_mask=True,
            image_id=True,
        )
        # Susceptibility distortion correction using T1w image
        sdc = epi_pipeline(
            str(self.base_dir),
            self.parameters["delete_cache"],
            name="SusceptibilityDistortionCorrection",
            ants_random_seed=self.parameters["random_seed"],
            use_double_precision=self.parameters["double_precision"],
        )

        # Remove bias correction from (Jeurissen et al., 2014)
        bias = npe.Node(
            mrtrix3.DWIBiasCorrect(
                use_ants=True,
                out_file="Jacobian_image_maths_thresh_merged_biascorr.nii.gz",
            ),
            name="RemoveBias",
        )
        # Compute b0 mask on corrected avg b0
        compute_avg_b0 = npe.Node(
            nutil.Function(
                input_names=["dwi_filename", "b_value_filename"],
                output_names=["out_b0_average"],
                function=compute_average_b0_task,
            ),
            name="ComputeB0Average",
        )
        compute_avg_b0.inputs.b_value_threshold = self.parameters["low_bval"]

        # Compute brain mask from reference b0
        mask_avg_b0 = npe.Node(fsl.BET(mask=True, robust=True), name="MaskB0")

        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=["image_id", "final_file"], function=print_end_pipeline
            ),
            name="WriteEndMessage",
        )

        self.connect(
            [
                (self.input_node, init_node, [(x, x) for x in self.get_input_fields()]),
                (
                    init_node,
                    prepare_b0,
                    [
                        (x, x)
                        for x in (
                            "dwi_filename",
                            "b_values_filename",
                            "b_vectors_filename",
                        )
                    ],
                ),
                (
                    init_node,
                    eddy_fsl,
                    [
                        (x, f"inputnode.{x}")
                        for x in (
                            "total_readout_time",
                            "phase_encoding_direction",
                            "image_id",
                        )
                    ],
                ),
                (
                    prepare_b0,
                    eddy_fsl,
                    [
                        ("out_b0_dwi_merge", "inputnode.dwi_filename"),
                        ("out_updated_bval", "inputnode.b_values_filename"),
                        ("out_updated_bvec", "inputnode.b_vectors_filename"),
                        ("out_reference_b0", "inputnode.reference_b0"),
                    ],
                ),
                # Magnetic susceptibility correction
                (init_node, sdc, [("t1w_filename", "inputnode.t1_filename")]),
                (
                    eddy_fsl,
                    sdc,
                    [("outputnode.out_corrected", "inputnode.dwi_filename")],
                ),
                (
                    eddy_fsl,
                    sdc,
                    [("outputnode.out_rotated_bvecs", "inputnode.b_vectors_filename")],
                ),
                # Bias correction
                (prepare_b0, bias, [("out_updated_bval", "in_bval")]),
                (
                    sdc,
                    bias,
                    [
                        ("outputnode.epi_corrected_dwi_image", "in_file"),
                        ("outputnode.rotated_b_vectors", "in_bvec"),
                    ],
                ),
                # Compute average b0 on corrected dataset (for brain mask extraction)
                (
                    prepare_b0,
                    compute_avg_b0,
                    [("out_updated_bval", "b_value_filename")],
                ),
                (bias, compute_avg_b0, [("out_file", "dwi_filename")]),
                # Compute b0 mask on corrected avg b0
                (compute_avg_b0, mask_avg_b0, [("out_b0_average", "in_file")]),
                # Print end message
                (init_node, print_end_message, [("image_id", "image_id")]),
                (mask_avg_b0, print_end_message, [("mask_file", "final_file")]),
                # Output node
                (bias, self.output_node, [("out_file", "preproc_dwi")]),
                (
                    sdc,
                    self.output_node,
                    [("outputnode.rotated_b_vectors", "preproc_bvec")],
                ),
                (prepare_b0, self.output_node, [("out_updated_bval", "preproc_bval")]),
                (mask_avg_b0, self.output_node, [("mask_file", "b0_mask")]),
            ]
        )



from typing import Optional

import click

from clinica import option
from clinica.iotools.converters import cli_param


@click.command(name="oasis-to-bids")
@cli_param.dataset_directory
@cli_param.clinical_data_directory
@cli_param.bids_directory
@option.global_option_group
@option.n_procs
def cli(
    dataset_directory: str,
    clinical_data_directory: str,
    bids_directory: str,
    n_procs: Optional[int] = None,
) -> None:
    """OASIS to BIDS converter.

    Convert the imaging and clinical data of OASIS (https://sites.wustl.edu/oasisbrains/),
    located in DATASET_DIRECTORY and CLINICAL_DATA_DIRECTORY respectively,
    to a BIDS dataset in the target BIDS_DIRECTORY.
    """
    from .oasis_to_bids import convert

    convert(
        dataset_directory,
        bids_directory,
        clinical_data_directory,
        n_procs=n_procs,
    )


if __name__ == "__main__":
    cli()



from pathlib import Path
from typing import Iterable

import pandas as pd

__all__ = [
    "read_clinical_data",
    "read_imaging_data",
    "intersect_data",
    "dataset_to_bids",
    "write_bids",
]


def read_clinical_data(clinical_data_directory: Path) -> dict[str, pd.DataFrame]:
    try:
        image_data_files = [f for f in clinical_data_directory.glob("*.csv")]
    except StopIteration:
        raise FileNotFoundError("Imaging collection file not found.")
    image_metadata = [pd.read_csv(image_path) for image_path in image_data_files]
    for i, df in enumerate(image_metadata):
        image_metadata[i] = df.set_index(df.axes[1][0])

    return {
        k: _get_df_based_on_index_name(image_metadata, index_name)
        for k, index_name in zip(
            ["pet", "mri", "subject", "pup", "adrc"],
            [
                "XNAT_PETSESSIONDATA ID",
                "MR ID",
                "Subject",
                "PUP_PUPTIMECOURSEDATA ID",
                "ADRC_ADRCCLINICALDATA ID",
            ],
        )
    }


def _get_df_based_on_index_name(
    image_metadata: list[pd.DataFrame], index_name: str
) -> pd.DataFrame:
    matching_dfs = [df for df in image_metadata if df.index.name == index_name]
    if len(matching_dfs) == 0:
        raise FileNotFoundError(f"Clinical data not found for {index_name}.")
    if len(matching_dfs) > 1:
        raise ValueError(f"Multiple data found for {index_name}.")
    return matching_dfs[0]


def read_imaging_data(imaging_data_directory: Path) -> pd.DataFrame:
    from clinica.iotools.bids_utils import StudyName, bids_id_factory

    source_path_series = pd.Series(
        _find_imaging_data(imaging_data_directory), name="source_path"
    )
    source_dir_series = source_path_series.apply(lambda x: x.parent).rename(
        "source_dir"
    )
    file_spec_series = source_path_series.apply(lambda x: x.parts[0]).rename("path")
    source_file_series = source_path_series.apply(
        lambda x: _identify_modality(x)
    ).rename("modality")
    source_run_series = source_path_series.apply(lambda x: _identify_runs(x)).rename(
        "run_number"
    )
    df_source = pd.concat(
        [
            source_path_series,
            file_spec_series,
            source_dir_series,
            file_spec_series.str.split("_", expand=True),
            source_file_series,
            source_run_series,
        ],
        axis=1,
    )
    df_source = (
        df_source.rename({0: "Subject", 1: "modality_2", 2: "Date"}, axis="columns")
        .drop_duplicates()
        .sort_values(by=["source_path"])
    )
    df_source = df_source.assign(
        participant_id=lambda df: df.Subject.apply(
            lambda x: bids_id_factory(StudyName.OASIS3).from_original_study_id(x)
        )
    )
    df_source["modality"] = df_source[["modality", "modality_2"]].apply(
        "_".join, axis=1
    )
    return df_source


def _identify_modality(source_path: Path) -> str:
    try:
        return source_path.name.split(".")[0].split("_")[-1]
    except Exception:
        return "nan"


def _identify_runs(source_path: Path) -> str:
    import re

    try:
        return re.search(r"run-\d+", str(source_path))[0]
    except Exception:
        return "run-01"


def _find_imaging_data(path_to_source_data: Path) -> Iterable[Path]:
    for image in path_to_source_data.rglob("*.nii.gz"):
        yield image.relative_to(path_to_source_data)


def intersect_data(
    df_source: pd.DataFrame, dict_df: dict
) -> tuple[pd.DataFrame, pd.DataFrame]:
    df_adrc = dict_df["adrc"].assign(
        session_id=lambda df: df.index.map(lambda x: x.split("_")[2])
    )
    df_adrc_small = df_adrc[df_adrc.session_id == "d0000"]
    df_adrc_small = df_adrc_small.merge(
        df_source["Subject"], how="inner", on="Subject"
    ).drop_duplicates()
    df_source = df_source.merge(
        df_adrc_small[["Subject", "ageAtEntry"]], how="inner", on="Subject"
    )
    df_source = df_source.assign(
        age=lambda df: (df["ageAtEntry"]) + df["Date"].str[1:].astype("float") / 365.25
    )
    df_subject_small = (
        dict_df["subject"]
        .merge(df_source["Subject"], how="inner", on="Subject")
        .drop_duplicates()
    )
    df_source = df_source.merge(
        dict_df["mri"]["Scanner"], how="left", left_on="path", right_on="MR ID"
    )
    df_small = df_subject_small.merge(df_adrc_small, how="inner", on="Subject")
    df_small = df_small.assign(participant_id=lambda df: "sub-" + df.Subject)
    df_source = df_source.assign(
        session=lambda df: (round(df["Date"].str[1:].astype("int") / (365.25 / 2)) * 6)
    )
    df_source = df_source.assign(
        ses=lambda df: df.session.apply(
            lambda x: f"ses-M{(5 - (len(str(x)))) * '0' + str(int(x))}"
        )
    )
    df_source = df_source.join(
        df_source.modality.map(
            {
                "dwi_MR": {"datatype": "dwi", "suffix": "dwi"},
                "T1w_MR": {"datatype": "anat", "suffix": "T1w"},
                "T2star_MR": {"datatype": "anat", "suffix": "T2starw"},
                "FLAIR_MR": {"datatype": "anat", "suffix": "FLAIR"},
                "pet_FDG": {"datatype": "pet", "suffix": "pet", "trc_label": "18FFDG"},
                "pet_PIB": {"datatype": "pet", "suffix": "pet", "trc_label": "11CPIB"},
                "pet_AV45": {
                    "datatype": "pet",
                    "suffix": "pet",
                    "trc_label": "18FAV45",
                },
            }
        ).apply(pd.Series)
    )
    if "trc_label" in df_source.columns:
        df_source = df_source.assign(
            filename=lambda df: df.apply(
                lambda x: f"{x.participant_id}/{x.ses}/{x.datatype}/"
                f"{x.participant_id}_{x.ses}"
                f"{'_trc-' + x.trc_label if pd.notna(x.trc_label) else ''}"
                f"_{x.run_number}_{x.suffix}.nii.gz",
                axis=1,
            )
        )
    else:
        df_source = df_source.assign(
            filename=lambda df: df.apply(
                lambda x: f"{x.participant_id}/{x.ses}/{x.datatype}/"
                f"{x.participant_id}_{x.ses}"
                f"_{x.run_number}_{x.suffix}.nii.gz",
                axis=1,
            )
        )
    df_adrc = df_adrc.merge(df_source["Subject"], how="inner", on="Subject")
    df_adrc = df_adrc.assign(
        session=lambda df: round(df["session_id"].str[1:].astype("int") / (364.25 / 2))
        * 6
    )
    df_adrc = df_adrc.drop_duplicates().set_index(["Subject", "session_id"])
    df_source = df_source.merge(
        df_adrc[
            [
                "session",
                "mmse",
                "cdr",
                "commun",
                "dx1",
                "homehobb",
                "judgment",
                "memory",
                "orient",
                "perscare",
                "sumbox",
                "apoe",
            ]
        ],
        how="left",
        on="session",
    )
    return df_source, df_small


def dataset_to_bids(
    df_source: pd.DataFrame, df_small: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    return (
        _build_participants_df(df_small),
        _build_sessions_df(df_source),
        _build_scans_df(df_source),
    )


def _build_participants_df(df_small: pd.DataFrame) -> pd.DataFrame:
    df_participants = (
        df_small[["participant_id", "ageAtEntry", "M/F", "Hand", "Education"]]
        .rename(
            columns={
                "ageAtEntry": "age",
                "M/F": "sex",
                "Hand": "handedness",
            }
        )
        .set_index("participant_id", verify_integrity=True)
    )
    return df_participants


def _build_sessions_df(df_source: pd.DataFrame) -> pd.DataFrame:
    df_session = (
        df_source[
            [
                "participant_id",
                "ses",
                "Date",
                "age",
                "mmse",
                "cdr",
                "commun",
                "dx1",
                "homehobb",
                "judgment",
                "memory",
                "orient",
                "perscare",
                "sumbox",
                "apoe",
            ]
        ]
        .rename(
            columns={
                "ses": "session_id",
                "Date": "source_session_id",
            }
        )
        .astype({"age": "int"})
        .drop_duplicates(subset=["participant_id", "session_id"])
        .set_index(["participant_id", "session_id"], verify_integrity=True)
    )
    return df_session


def _build_scans_df(df_source: pd.DataFrame) -> pd.DataFrame:
    df_scan = (
        df_source[["filename", "source_dir"]]
        .drop_duplicates(subset=["filename"])
        .set_index("filename", verify_integrity=True)
    )
    return df_scan


def _install_bids(sourcedata_dir: Path, bids_filename: Path) -> None:
    from fsspec.implementations.local import LocalFileSystem

    fs = LocalFileSystem(auto_mkdir=True)

    source_file = fs.open(fs.ls(sourcedata_dir)[0], mode="rb")
    target_file = fs.open(bids_filename, mode="wb")

    with source_file as sf, target_file as tf:
        tf.write(sf.read())

    source_basename = Path(Path(Path(fs.ls(sourcedata_dir)[0]).stem).stem)
    target_basename = Path(bids_filename.stem).stem

    # The following part adds the sidecar files related to the nifti with the same name: it can be tsv or json files.
    # It may or may not be used, since there might not be any sidecars.
    sidecar_dir = sourcedata_dir.parent / "BIDS"
    for source_sidecar in sidecar_dir.rglob(f"{source_basename}*"):
        target_sidecar = (bids_filename.parent / target_basename).with_name(
            f"{target_basename}{source_sidecar.suffix}"
        )
        source_file = fs.open(source_sidecar, mode="rb")
        target_file = fs.open(target_sidecar, mode="wb")
        with source_file as sf, target_file as tf:
            tf.write(sf.read())


def write_bids(
    to: Path,
    participants: pd.DataFrame,
    sessions: pd.DataFrame,
    scans: pd.DataFrame,
    dataset_directory: Path,
) -> list[str]:
    from fsspec.implementations.local import LocalFileSystem

    from clinica.iotools.bids_dataset_description import BIDSDatasetDescription
    from clinica.iotools.bids_utils import StudyName, write_to_tsv

    fs = LocalFileSystem(auto_mkdir=True)

    with fs.transaction:
        with fs.open(to / "dataset_description.json", "w") as dataset_description_file:
            BIDSDatasetDescription(name=StudyName.OASIS3).write(
                to=dataset_description_file
            )
        with fs.open(to / "participants.tsv", "w") as participant_file:
            write_to_tsv(participants, participant_file)
        for participant_id, sessions_group in sessions.groupby("participant_id"):
            sessions_group = sessions_group.droplevel("participant_id")
            sessions_filepath = to / participant_id / f"{participant_id}_sessions.tsv"
            with fs.open(sessions_filepath, "w") as sessions_file:
                write_to_tsv(sessions_group, sessions_file)

    for filename, metadata in scans.iterrows():
        path = dataset_directory / metadata.source_dir
        if _extract_suffix_from_filename(str(filename)) != "nan":
            _install_bids(sourcedata_dir=path, bids_filename=to / filename)
    return scans.index.to_list()


def _extract_suffix_from_filename(filename: str) -> str:
    return filename.split("_")[-1].split(".")[0]



from ._centering import (
    center_all_nifti,
    center_nifti_origin,
    check_relative_volume_location_in_world_coordinate_system,
    check_volume_location_in_world_coordinate_system,
)
from ._files import create_subs_sess_list, write_list_of_files
from ._merging import create_merge_file
from ._missing import compute_missing_mods, compute_missing_processing

__all__ = [
    "create_merge_file",
    "center_nifti_origin",
    "center_all_nifti",
    "compute_missing_mods",
    "compute_missing_processing",
    "create_subs_sess_list",
    "write_list_of_files",
    "check_volume_location_in_world_coordinate_system",
    "check_relative_volume_location_in_world_coordinate_system",
]



import click

from clinica.pipelines.engine import clinica_pipeline


@clinica_pipeline
@click.command(
    "deeplearning-prepare-data",
    deprecated=True,
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
    },
)
def cli(*args, **kwargs):
    """Prepare data generated by Clinica for PyTorch with tensor extraction."""
    click.echo(
        "Functionalities provided by this pipeline have been moved to the `clinicadl` package."
        "\n\n"
        "See: https://clinicadl.readthedocs.io/en/stable/Preprocessing/Extract/"
    )



from typing import List

import numpy as np
from pydra.mark import annotate, task


@task
@annotate({"return": {"g_fisher_tensor": np.array, "output_fisher_tensor": str}})
def obtain_g_fisher_tensor_task(dartel_input, FWHM):
    """Pydra task for obtaining g fisher tensor.

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils.obtain_g_fisher_tensor`.
    """
    from clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils import (
        obtain_g_fisher_tensor,
    )

    return obtain_g_fisher_tensor(dartel_input, FWHM)


@task
@annotate({"return": {"t_step": np.array, "output_data": str}})
def obtain_time_step_estimation_task(dartel_input, FWHM, g):
    """Pydra task for computing time step estimation.

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils.obtain_time_step_estimation`.
    """
    from clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils import (
        obtain_time_step_estimation,
    )

    return obtain_time_step_estimation(dartel_input, FWHM, g)


@task
@annotate({"return": {"regularized": str}})
def heat_solver_equation_task(input_image, g, FWHM, t_step, dartel_input):
    """Pydra task for computing heat solver equation.

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils.heat_solver_equation`.
    """
    from clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils import (
        heat_solver_equation,
    )

    return heat_solver_equation(input_image, g, FWHM, t_step, dartel_input)



from pathlib import Path
from typing import List

from clinica.pipelines.engine import Pipeline


class T1FreeSurferTemplate(Pipeline):
    """FreeSurfer Longitudinal template class.

    Returns:
        A clinica pipeline object containing the T1FreeSurferTemplate pipeline.
    """

    @staticmethod
    def get_processed_images(
        caps_directory: Path, subjects: List[str], sessions: List[str]
    ) -> List[str]:
        import re

        from clinica.utils.input_files import T1_FS_T_DESTRIEUX
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.longitudinal import get_long_id
        from clinica.utils.participant import get_unique_subjects

        [list_participant_id, list_list_session_ids] = get_unique_subjects(
            subjects, sessions
        )
        list_long_id = [
            get_long_id(list_session_ids) for list_session_ids in list_list_session_ids
        ]

        image_ids: List[str] = []
        if caps_directory.is_dir():
            t1_freesurfer_files, _ = clinica_file_reader(
                list_participant_id,
                list_long_id,
                caps_directory,
                T1_FS_T_DESTRIEUX,
                False,
            )
            image_ids = [
                re.search(r"(sub-[a-zA-Z0-9]+)_(long-[a-zA-Z0-9]+)", file).group()
                for file in t1_freesurfer_files
            ]
        return image_ids

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        pass

    def _check_custom_dependencies(self) -> None:
        """Check dependencies that cannot be listed in `info.json`."""
        pass

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Notes
        -----
        The list of inputs of the T1FreeSurferTemplate pipeline is:
            * participant_id (str): Participant ID (e.g. "sub-CLNC01")
            * list_session_ids (List[str]: List of session IDs associated to `participant_id`
                (e.g. ["ses-M000"] or ["ses-M000", "ses-M018", "ses-M036"])

        Returns
        -------
        list of str :
            A list of (string) input fields name.
        """
        return ["participant_id", "list_session_ids"]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Notes
        -----
        The list of inputs of the T1FreeSurferTemplate pipeline is:
            * image_id (str): Image ID (e.g. "sub-CLNC01_long-M000M018")

        Returns
        -------
        list of str :
            A list of (string) output fields name.
        """
        return ["image_id"]

    def _build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.pipelines.anatomical.freesurfer.longitudinal.utils import (
            save_part_sess_long_ids_to_tsv,
        )
        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.filemanip import extract_subjects_sessions_from_filename
        from clinica.utils.input_files import T1_FS_DESTRIEUX
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.longitudinal import (
            get_long_id,
            get_participants_long_id,
            read_sessions,
        )
        from clinica.utils.participant import (
            get_unique_subjects,
            unique_subjects_sessions_to_subjects_sessions,
        )
        from clinica.utils.stream import cprint

        from .utils import extract_participant_long_ids_from_filename

        output_ids = self.get_processed_images(
            self.caps_directory, self.subjects, self.sessions
        )
        (
            processed_participants,
            processed_long_sessions,
        ) = extract_participant_long_ids_from_filename(output_ids)
        if len(processed_participants) > 0:
            cprint(
                msg=(
                    f"Clinica found {len(processed_participants)} participant(s) "
                    "already processed in CAPS directory:"
                ),
                lvl="warning",
            )
            for p_id, l_id in zip(processed_participants, processed_long_sessions):
                cprint(f"{p_id} | {l_id}", lvl="warning")
            if self.overwrite_caps:
                output_folder = "<CAPS>/subjects/<participant_id>/<long_id>/freesurfer_unbiased_template/"
                cprint(
                    f"Output folders in {output_folder} will be recreated.",
                    lvl="warning",
                )
            else:
                cprint("Participant(s) will be ignored by Clinica.", lvl="warning")
                input_ids = [
                    f"{p_id}_{s_id}" for p_id, s_id in zip(self.subjects, self.sessions)
                ]
                processed_sessions_per_participant = [
                    read_sessions(self.caps_directory, p_id, l_id)
                    for (p_id, l_id) in zip(
                        processed_participants, processed_long_sessions
                    )
                ]
                participants, sessions = unique_subjects_sessions_to_subjects_sessions(
                    processed_participants, processed_sessions_per_participant
                )
                processed_ids = [
                    f"{p_id}_{s_id}" for p_id, s_id in zip(participants, sessions)
                ]
                to_process_ids = list(set(input_ids) - set(processed_ids))
                self.subjects, self.sessions = extract_subjects_sessions_from_filename(
                    to_process_ids
                )
        try:
            clinica_file_reader(
                self.subjects, self.sessions, self.caps_directory, T1_FS_DESTRIEUX
            )
        except ClinicaException as e:
            err_msg = (
                "Clinica faced error(s) while trying to read files in your CAPS directory.\n"
                + str(e)
            )
            raise ClinicaCAPSError(err_msg)

        long_ids = get_participants_long_id(self.subjects, self.sessions)
        save_part_sess_long_ids_to_tsv(
            self.subjects, self.sessions, long_ids, self.base_dir / self.name
        )
        [list_participant_id, list_list_session_ids] = get_unique_subjects(
            self.subjects, self.sessions
        )
        list_long_id = [
            get_long_id(list_session_ids) for list_session_ids in list_list_session_ids
        ]

        def print_images_to_process(
            unique_part_list: list[str],
            per_part_session_list: list[str],
            list_part_long_id: list[str],
        ):
            cprint(
                f"The pipeline will be run on the following {len(unique_part_list)} participant(s):"
            )
            for part_id, list_sess_id, list_id in zip(
                unique_part_list, per_part_session_list, list_part_long_id
            ):
                sessions_participant = ", ".join(s_id for s_id in list_sess_id)
                cprint(f"\t{part_id} | {sessions_participant} | {list_id}")

        if len(self.subjects):
            # TODO: Generalize long IDs to the message display
            print_images_to_process(
                list_participant_id, list_list_session_ids, list_long_id
            )
            cprint(
                f"List available in {self.base_dir / self.name / 'participants.tsv'}"
            )
            cprint("The pipeline will last approximately 10 hours per participant.")

        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                ("participant_id", list_participant_id),
                ("list_session_ids", list_list_session_ids),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        self.connect(
            [
                (read_node, self.input_node, [("participant_id", "participant_id")]),
                (
                    read_node,
                    self.input_node,
                    [("list_session_ids", "list_session_ids")],
                ),
            ]
        )

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from .tasks import save_to_caps_task

        save_to_caps = npe.Node(
            interface=nutil.Function(
                input_names=[
                    "source_dir",
                    "image_id",
                    "list_session_ids",
                    "caps_dir",
                    "overwrite_caps",
                ],
                output_names=["image_id"],
                function=save_to_caps_task,
            ),
            name="SaveToCaps",
        )
        save_to_caps.inputs.source_dir = str(self.base_dir / self.name / "ReconAll")
        save_to_caps.inputs.caps_dir = str(self.caps_directory)
        save_to_caps.inputs.overwrite_caps = False

        self.connect(
            [
                (
                    self.input_node,
                    save_to_caps,
                    [("list_session_ids", "list_session_ids")],
                ),
                (self.output_node, save_to_caps, [("image_id", "image_id")]),
            ]
        )

    def _build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from .tasks import init_input_node_task, move_subjects_dir_to_source_dir_task
        from .utils import run_recon_all_base

        init_input = npe.Node(
            interface=nutil.Function(
                input_names=[
                    "caps_dir",
                    "participant_id",
                    "list_session_ids",
                    "output_dir",
                ],
                output_names=["image_id", "subjects_dir", "flags"],
                function=init_input_node_task,
            ),
            name="0-InitPipeline",
        )
        init_input.inputs.caps_dir = str(self.caps_directory)
        init_input.inputs.output_dir = str(self.base_dir / self.name / "ReconAll")

        recon_all = npe.Node(
            interface=nutil.Function(
                input_names=["subjects_dir", "subject_id", "flags", "directive"],
                output_names=["subject_id"],
                function=run_recon_all_base,
            ),
            name="1-SegmentationReconAll",
        )
        recon_all.inputs.directive = "-all"

        move_subjects_dir = npe.Node(
            interface=nutil.Function(
                input_names=["subjects_dir", "source_dir", "subject_id"],
                output_names=["subject_id"],
                function=move_subjects_dir_to_source_dir_task,
            ),
            name="2-MoveSubjectsDir",
        )
        move_subjects_dir.inputs.source_dir = str(
            self.base_dir / self.name / "ReconAll"
        )

        connections = [
            (self.input_node, init_input, [("participant_id", "participant_id")]),
            (
                self.input_node,
                init_input,
                [("list_session_ids", "list_session_ids")],
            ),
            (
                init_input,
                recon_all,
                [
                    ("subjects_dir", "subjects_dir"),
                    ("image_id", "subject_id"),
                    ("flags", "flags"),
                ],
            ),
            (init_input, move_subjects_dir, [("subjects_dir", "subjects_dir")]),
            (recon_all, move_subjects_dir, [("subject_id", "subject_id")]),
            (move_subjects_dir, self.output_node, [("subject_id", "image_id")]),
        ]
        self.connect(connections)



import typing as ty
from pathlib import Path

import pydra
from nipype.interfaces.ants import ApplyTransforms, Registration, RegistrationSynQuick
from pydra.engine import Workflow
from pydra.tasks.nipype1.utils import Nipype1Task

from clinica.pipelines.pet.utils import get_suvr_mask
from clinica.pydra.engine import clinica_io
from clinica.pydra.pet_linear.tasks import (
    concatenate_transforms_task,
    crop_nifti_task,
    suvr_normalization_task,
)
from clinica.pydra.tasks import download_mni_template_2009c

IMAGE_DIMENSION = 3


@clinica_io
def build_core_workflow(name: str = "core", parameters: dict = {}) -> Workflow:
    """Build the core workflow for the PET Linear pipeline.

    Parameters
    ----------
    name : str, optional
        The name of the workflow. Default="core".

    parameters : dict, optional
        Optional dictionary of parameters to be used within the workflow.
        Default={}.

    Returns
    -------
    wf : Workflow
        The core workflow.
    """
    input_spec = pydra.specs.SpecInfo(
        name="Input",
        fields=[
            ("_graph_checksums", ty.Any),
            ("T1w", str, {"mandatory": True}),
            ("pet", dict, {"tracer": parameters["acq_label"]}, {"mandatory": True}),
            ("t1w_to_mni", dict, {}, {"mandatory": True}),
        ],
        bases=(pydra.specs.BaseSpec,),
    )

    ref_mask = get_suvr_mask(parameters["suvr_reference_region"])

    wf = Workflow(
        name,
        input_spec=input_spec,
    )

    wf.split(("pet", "T1w", "t1w_to_mni"))

    wf.add(download_mni_template_2009c(name="download_mni_template"))

    # RegistrationSynQuick by ANTS.
    ants_registration = Nipype1Task(
        name="ants_registration",
        interface=RegistrationSynQuick(),
    )
    ants_registration.inputs.dimension = IMAGE_DIMENSION
    ants_registration.inputs.transform_type = "r"
    ants_registration.inputs.fixed_image = wf.lzin.T1w
    ants_registration.inputs.moving_image = wf.lzin.pet
    wf.add(ants_registration)

    # Concatenate
    wf.add(
        concatenate_transforms_task(
            name="concatenate_transforms",
            interface=concatenate_transforms_task,
            pet_to_t1w_tranform=wf.ants_registration.lzout.out_matrix,
            t1w_to_mni_tranform=wf.lzin.t1w_to_mni,
        )
    )

    # ApplyTransforms by ANTS. PET to MRI
    apply_transform_pet_to_mni = Nipype1Task(
        name="apply_transform_pet_to_mni",
        interface=ApplyTransforms(),
    )
    apply_transform_pet_to_mni.inputs.dimension = IMAGE_DIMENSION
    apply_transform_pet_to_mni.inputs.reference_image = (
        wf.download_mni_template.lzout.mni_template_file
    )
    apply_transform_pet_to_mni.inputs.input_image = wf.lzin.pet
    apply_transform_pet_to_mni.inputs.transforms = (
        wf.concatenate_transforms.lzout.transforms_list
    )
    wf.add(apply_transform_pet_to_mni)

    # Normalize the image
    ants_registration_t1w_to_mni = Nipype1Task(
        name="ants_registration_t1w_to_mni",
        interface=Registration(),
    )
    ants_registration_t1w_to_mni.inputs.metric = ["MI"]
    ants_registration_t1w_to_mni.inputs.metric_weight = [1.0]
    ants_registration_t1w_to_mni.inputs.transforms = ["SyN"]
    ants_registration_t1w_to_mni.inputs.transform_parameters = [(0.1, 3, 0)]
    ants_registration_t1w_to_mni.inputs.dimension = IMAGE_DIMENSION
    ants_registration_t1w_to_mni.inputs.shrink_factors = [[8, 4, 2]]
    ants_registration_t1w_to_mni.inputs.smoothing_sigmas = [[3, 2, 1]]
    ants_registration_t1w_to_mni.inputs.sigma_units = ["vox"]
    ants_registration_t1w_to_mni.inputs.number_of_iterations = [[200, 50, 10]]
    ants_registration_t1w_to_mni.inputs.convergence_threshold = [1e-05]
    ants_registration_t1w_to_mni.inputs.convergence_window_size = [10]
    ants_registration_t1w_to_mni.inputs.radius_or_number_of_bins = [32]
    ants_registration_t1w_to_mni.inputs.winsorize_lower_quantile = 0.005
    ants_registration_t1w_to_mni.inputs.winsorize_upper_quantile = 0.995
    ants_registration_t1w_to_mni.inputs.collapse_output_transforms = True
    ants_registration_t1w_to_mni.inputs.use_histogram_matching = False
    ants_registration_t1w_to_mni.inputs.verbose = True
    ants_registration_t1w_to_mni.inputs.fixed_image = (
        wf.download_mni_template.lzout.mni_template_file
    )
    ants_registration_t1w_to_mni.inputs.moving_image = wf.lzin.T1w
    wf.add(ants_registration_t1w_to_mni)

    apply_transform_non_linear = Nipype1Task(
        name="apply_transform_non_linear",
        interface=ApplyTransforms(),
    )
    apply_transform_non_linear.inputs.dimension = IMAGE_DIMENSION
    apply_transform_non_linear.inputs.reference_image = (
        wf.download_mni_template.lzout.mni_template_file
    )
    apply_transform_non_linear.inputs.transforms = (
        wf.ants_registration_t1w_to_mni.lzout.reverse_forward_transforms
    )
    apply_transform_non_linear.inputs.input_image = (
        wf.apply_transform_pet_to_mni.lzout.output_image
    )
    wf.add(apply_transform_non_linear)

    wf.add(
        suvr_normalization_task(
            name="suvr_normalization",
            interface=suvr_normalization_task,
            input_img=wf.apply_transform_pet_to_mni.lzout.output_image,
            norm_img=wf.apply_transform_non_linear.lzout.output_image,
            ref_mask=ref_mask,
        )
    )

    output_connections = [
        ("affine_mat", wf.ants_registration.lzout.out_matrix),
        ("suvr_pet", wf.suvr_normalization.lzout.output_img),
    ]

    # Crop image
    if not parameters["uncropped_image"]:
        wf.add(
            crop_nifti_task(
                name="crop_nifti",
                interface=crop_nifti_task,
                input_img=wf.suvr_normalization.lzout.output_img,
                output_dir=Path.cwd(),
            )
        )

        output_connections += [
            ("outfile_crop", wf.crop_nifti.lzout.output_img),
        ]

    # Optional argument
    if parameters["save_PETinT1w"]:
        apply_transform_pet_to_t1w = Nipype1Task(
            name="apply_transform_pet_to_t1w",
            interface=ApplyTransforms(),
        )
        apply_transform_pet_to_t1w.inputs.dimension = IMAGE_DIMENSION
        apply_transform_pet_to_t1w.inputs.input_image = wf.lzin.pet
        apply_transform_pet_to_t1w.inputs.reference_image = wf.lzin.T1w
        apply_transform_pet_to_t1w.inputs.transforms = (
            wf.ants_registration.lzout.out_matrix
        )

        wf.add(apply_transform_pet_to_t1w)

        output_connections += [
            ("PETinT1w", wf.apply_transform_pet_to_t1w.lzout.output_image),
        ]

    wf.set_output(output_connections)

    return wf



"""This module contains Nipype tasks used by the DWIPreprocessingUsingT1 pipeline.

Nipype tasks must be 'self-contained' such that only primitive type hints can be
used. The tasks are simple wrappers around a properly implemented Python function.
"""


def prepare_reference_b0_task(
    dwi_filename: str,
    b_values_filename: str,
    b_vectors_filename: str,
    b_value_threshold: float = 5.0,
    working_directory=None,
) -> tuple:
    """Task called be Nipype to execute 'prepare_reference_b0'."""
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.t1.utils import prepare_reference_b0
    from clinica.pipelines.dwi.utils import DWIDataset

    if working_directory:
        working_directory = Path(working_directory)

    b0_reference_filename, reference_dwi_dataset = prepare_reference_b0(
        DWIDataset(
            dwi=dwi_filename,
            b_values=b_values_filename,
            b_vectors=b_vectors_filename,
        ),
        b_value_threshold,
        working_directory,
    )

    return str(b0_reference_filename), *(str(_) for _ in reference_dwi_dataset)


def change_itk_transform_type_task(input_affine_file: str) -> str:
    """Task called be Nipype to execute 'change_itk_transform_type'."""
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.t1.utils import change_itk_transform_type

    return str(change_itk_transform_type(Path(input_affine_file)))


def rotate_b_vectors_task(
    b_vectors_filename: str,
    matrix_filenames: list,
    output_dir: str = None,
) -> str:
    """Task called be Nipype to execute 'change_itk_transform_type'."""
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.t1.utils import rotate_b_vectors

    return str(
        rotate_b_vectors(
            Path(b_vectors_filename),
            matrix_filenames,
            Path(output_dir) if output_dir else None,
        )
    )


def broadcast_matrix_filename_to_match_b_vector_length_task(
    matrix_filename: str, b_vectors_filename: str
) -> list:
    """Task called be Nipype to execute 'broadcast_matrix_filename_to_match_b_vector_length'."""
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.t1.utils import (
        broadcast_matrix_filename_to_match_b_vector_length,
    )

    return [
        str(file_path)
        for file_path in broadcast_matrix_filename_to_match_b_vector_length(
            Path(matrix_filename), Path(b_vectors_filename)
        )
    ]


def rename_into_caps_task(
    dwi_filename: str,
    dwi_preproc_filename: str,
    b_values_preproc_filename: str,
    b_vectors_preproc_filename: str,
    b0_brain_mask_filename: str,
) -> tuple:
    """Task called be Nipype to execute 'rename_into_caps'."""
    from pathlib import Path

    from clinica.pipelines.dwi.preprocessing.t1.utils import rename_into_caps

    return rename_into_caps(
        Path(dwi_filename),
        Path(dwi_preproc_filename),
        Path(b_values_preproc_filename),
        Path(b_vectors_preproc_filename),
        Path(b0_brain_mask_filename),
    )


def merge_nifti_images_in_time_dimension_task(image1: str, image2: str) -> str:
    """Merges the two provided volumes in the time (4th) dimension."""
    from clinica.utils.image import merge_nifti_images_in_time_dimension

    return str(merge_nifti_images_in_time_dimension((image1, image2)))



from typing import Optional

import click

import clinica.pydra.engine_utils as pydra_utils
import clinica.pydra.statistics_volume_correction.pipeline as pydra_statistics_volume_correction
from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "pydra-statistics-volume-correction"


@clinica_pipeline
@click.command(name=pipeline_name)
@cli_param.argument.caps_directory
@click.argument("t_map", type=str)
@click.argument("height_threshold", type=float)
@click.argument("FWEp", type=float)
@click.argument("FDRp", type=float)
@click.argument("FWEc", type=float)
@click.argument("FDRc", type=float)
@cli_param.option_group.pipeline_specific_options
@cli_param.option_group.option(
    "-nc",
    "--n_cuts",
    default=8,
    show_default=True,
    help="Number of cuts along each direction",
)
@cli_param.option_group.common_pipelines_options
@cli_param.option.working_directory
@option.global_option_group
@option.n_procs
def cli(
    caps_directory: str,
    t_map: str,
    height_threshold: float,
    fwep: float,
    fdrp: float,
    fwec: float,
    fdrc: float,
    n_cuts: int = 8,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Statistical correction of statistics-volume pipeline.

       T_MAP is the name of the T statistic map used for the correction.

       HEIGHT_THRESHOLD is the T value corresponding to an uncorrected p-value of 0.001.

       FWEp is the Family-Wise Error peak threshold.

       FDRp is the False Discovery Rate peak threshold.

       Finally, FWEc and FDRc stand for the respective correction parameters.

    Prerequisite: You have to perform the statistics-volume pipeline before running this pipeline.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/Stats_Volume/
    """

    parameters = {
        "t_map": t_map,
        "height_threshold": height_threshold,
        "FWEp": fwep,
        "FDRp": fdrp,
        "FWEc": fwec,
        "FDRc": fdrc,
        "n_cuts": n_cuts,
    }

    pipeline = pydra_statistics_volume_correction.build_core_workflow(
        name="statistics-volume-correction-pydra",
        input_dir=None,
        output_dir=caps_directory,
        parameters=parameters,
    )
    pydra_utils.run(pipeline, n_procs)


if __name__ == "__main__":
    cli()



from enum import Enum
from pathlib import Path
from typing import Iterable, List, Optional

import pandas as pd

from clinica.utils.pet import ReconstructionMethod, Tracer
from clinica.utils.stream import cprint

__all__ = [
    "get_subjects_list",
    "ADNIModality",
    "ADNIModalityConverter",
    "ADNIStudy",
    "correct_diagnosis_sc_adni3",
    "create_adni_sessions_dict",
    "create_adni_scans_files",
    "paths_to_bids",
    "load_clinical_csv",
]


class ADNIStudy(str, Enum):
    """Possible versions of ADNI studies."""

    ADNI1 = "ADNI1"
    ADNI2 = "ADNI2"
    ADNI3 = "ADNI3"
    ADNIGO = "ADNIGO"


def _define_subjects_list(
    source_dir: Path,
    subjs_list_path: Optional[Path] = None,
) -> List[str]:
    import re

    from clinica.utils.stream import cprint

    if subjs_list_path:
        cprint("Loading a subjects lists provided by the user...")
        return subjs_list_path.read_text().splitlines()

    cprint(f"Using the subjects contained in the ADNI dataset at {source_dir}")
    rgx = re.compile(r"\d{3}_S_\d{4}")
    return list(filter(rgx.fullmatch, [folder.name for folder in source_dir.iterdir()]))


def _check_subjects_list(
    subjs_list: List[str],
    clinical_dir: Path,
) -> List[str]:
    from copy import copy

    from clinica.utils.stream import cprint

    subjs_list_copy = copy(subjs_list)
    adni_merge = load_clinical_csv(clinical_dir, "ADNIMERGE")
    # Check that there are no errors in subjs_list given by the user
    for subj in subjs_list_copy:
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        if len(adnimerge_subj) == 0:
            cprint(
                msg=f"Subject with PTID {subj} does not have corresponding clinical data."
                f"Please check your subjects list or directory.",
                lvl="warning",
            )
            subjs_list.remove(subj)
    del subjs_list_copy

    if not subjs_list:
        cprint(f"Processing an empty list of subjects.", lvl="warning")

    return subjs_list


def get_subjects_list(
    source_dir: Path,
    clinical_dir: Path,
    subjs_list_path: Optional[Path] = None,
) -> List[str]:
    return _check_subjects_list(
        _define_subjects_list(source_dir, subjs_list_path), clinical_dir
    )


class ADNIModality(str, Enum):
    """Possible modalities supported by the ADNI-to-BIDS converter.

    These are the modalities exposed to the user. There is not a
    one-to-one relationship with the modality converters. That is,
    some modalities, like PET_AMYLOID, are associated with multiple
    converters, while others are associated with only one converter.
    For this reason, the ADNIModalityConverter enumeration exists,
    and there is a mapping between a ADNIModality and an iterable of
    ADNIModalityConverter variants.
    """

    T1 = "T1"
    PET_FDG = "PET_FDG"
    PET_AMYLOID = "PET_AMYLOID"
    PET_TAU = "PET_TAU"
    DWI = "DWI"
    FLAIR = "FLAIR"
    FMRI = "fMRI"
    FMAP = "FMAP"


class ADNIModalityConverter(str, Enum):
    """Possible modality converters for ADNI.

    These are not exposed to the user. However, there is a one-to-one
    relationship with the modality converters. That is, each variant
    has a corresponding converter.
    """

    T1 = "T1"
    PET_FDG = "PET_FDG"
    PET_FDG_UNIFORM = "PET_FDG_UNIFORM"
    PET_PIB = "PET_PIB"
    PET_AV45 = "PET_AV45"
    PET_TAU = "PET_TAU"
    DWI = "DWI"
    FLAIR = "FLAIR"
    FMRI = "fMRI"
    FMAP = "FMAP"


def correct_diagnosis_sc_adni3(
    clinical_data_dir: Path, participants_df: pd.DataFrame
) -> pd.DataFrame:
    """Add missing diagnosis at screening in ADNI3 due to changes in DX_bl.

    See https://groups.google.com/g/adni-data/c/whYhKafN_0Q.
    The identification of SMC, EMCI, LMCI is lost in ADNI3.

    Args:
        clinical_data_dir: path to the input folder containing clinical data.
        participants_df: DataFrame containing metadata at the participant level.

    Returns:
        Corrected participants_df.
    """
    diagnosis_dict = {1: "CN", 2: "MCI", 3: "AD"}
    # DXSUM_PDXCONV_ADNIALL has been renamed to DXSUM_PDXCONV
    # this ensures old ADNI downloads still work with recent versions of Clinica
    try:
        dxsum_df = load_clinical_csv(
            clinical_data_dir, "DXSUM_PDXCONV_ADNIALL"
        ).set_index(["PTID", "VISCODE2"])
    except OSError:
        dxsum_df = load_clinical_csv(clinical_data_dir, "DXSUM_PDXCONV").set_index(
            ["PTID", "VISCODE2"]
        )
    missing_sc = participants_df[
        participants_df.original_study == ADNIStudy.ADNI3.value
    ]
    participants_df.set_index("alternative_id_1", drop=True, inplace=True)
    for alternative_id in missing_sc.alternative_id_1.values:
        try:
            diagnosis_sc = diagnosis_dict[
                dxsum_df.loc[(alternative_id, "sc"), "DIAGNOSIS"].values[0]
            ]
            participants_df.loc[alternative_id, "diagnosis_sc"] = diagnosis_sc
        except KeyError:
            cprint(
                msg=f"Unknown screening diagnosis for subject {alternative_id}.",
                lvl="warning",
            )
            participants_df.loc[alternative_id, "diagnosis_sc"] = "n/a"

    participants_df.reset_index(inplace=True, drop=False)
    return participants_df


def _write_adni_sessions_tsv(
    df_subj_sessions: pd.DataFrame, bids_subjs_paths: list[Path]
):
    """Write the result of method create_session_dict into several TSV files.

    Args:
        df_subj_sessions: global dataframe containing clinical sessions data for all subjects
        bids_subjs_paths: a list with the path to all bids subjects
    """

    df_subj_sessions["adas_memory"] = (
        df_subj_sessions["adas_Q1"]
        + df_subj_sessions["adas_Q4"]
        + df_subj_sessions["adas_Q7"]
        + df_subj_sessions["adas_Q8"]
        + df_subj_sessions["adas_Q9"]
    )  # / 45
    df_subj_sessions["adas_language"] = (
        df_subj_sessions["adas_Q2"]
        + df_subj_sessions["adas_Q5"]
        + df_subj_sessions["adas_Q10"]
        + df_subj_sessions["adas_Q11"]
        + df_subj_sessions["adas_Q12"]
    )  # / 25
    df_subj_sessions["adas_praxis"] = (
        df_subj_sessions["adas_Q3"] + df_subj_sessions["adas_Q6"]
    )  # / 10
    df_subj_sessions["adas_concentration"] = df_subj_sessions["adas_Q13"]  # / 5

    df_subj_sessions = df_subj_sessions.fillna("n/a")

    # compute the amyloid and ptau status
    df_subj_sessions["a_stat"] = df_subj_sessions.apply(
        lambda x: _compute_amyloid_status(
            x[["adni_av45", "adni_pib", "adni_abeta"]].apply(
                pd.to_numeric, errors="coerce"
            )
        ),
        axis=1,
    )
    df_subj_sessions["tau_stat"] = df_subj_sessions.apply(
        lambda x: _compute_ptau_status(
            x[["adni_ptau"]].apply(pd.to_numeric, errors="coerce")
        ),
        axis=1,
    )
    for subject_path in bids_subjs_paths:
        subject_path.mkdir(parents=True, exist_ok=True)
        if subject_path.name == "conversion_info":
            continue
        df_tmp = df_subj_sessions[df_subj_sessions["RID"] == subject_path.name[12:]]
        df_tmp.to_csv(
            subject_path / f"{subject_path.name}_sessions.tsv",
            sep="\t",
            index=False,
            encoding="utf-8",
        )


def _compute_amyloid_status(row: pd.DataFrame) -> str:
    if (
        pd.isnull(row["adni_av45"])
        and pd.isnull(row["adni_pib"])
        and isinstance(row["adni_abeta"], float)
    ):
        return "Au"
    if row["adni_av45"] > 1.1 or row["adni_pib"] > 1.5 or row["adni_abeta"] < 192:
        return "A+"
    if (
        (pd.isnull(row["adni_av45"]) or row["adni_av45"] < 1.1)
        and (pd.isnull(row["adni_pib"]) or row["adni_pib"] < 1.5)
        and (isinstance(row["adni_abeta"], float) or row["adni_abeta"] > 192)
    ):
        return "A-"
    return "n/a"


def _compute_ptau_status(row: pd.DataFrame) -> str:
    if pd.isnull(row["adni_ptau"]):
        return "Tu"
    if row["adni_ptau"] > 23:
        return "T+"
    return "T-"


def _filter_subj_bids(
    df_files: pd.DataFrame, location: str, bids_ids: list[str]
) -> pd.DataFrame:
    from clinica.iotools.bids_utils import remove_space_and_symbols

    # Depending on the file that needs to be open, identify and
    # preprocess the column that contains the subjects ids.
    # todo : use id class here ?
    bids_ids = [x[8:] for x in bids_ids if "sub-ADNI" in x]
    if location == "ADNIMERGE.csv":
        df_files["RID"] = df_files["PTID"].apply(
            lambda x: remove_space_and_symbols(x)[4:]
        )
    else:
        df_files["RID"] = df_files["RID"].apply(lambda x: _pad_id(x))
    return df_files.loc[df_files["RID"].isin([x[4:] for x in bids_ids])]


def _update_age(row: pd.DataFrame) -> float:
    """Update age with time passed since bl to current visit"""
    from datetime import datetime

    if row["session_id"] != "ses-M000":
        exam_date = datetime.strptime(row["EXAMDATE"], "%Y-%m-%d")
        exam_date_baseline = datetime.strptime(row["EXAMDATE_bl"], "%Y-%m-%d")
        delta = exam_date - exam_date_baseline
        return round(
            float(row["AGE"]) + (delta.days / 365.25),
            1,
        )
    return row["AGE"]


def _pad_id(ref_id: str) -> str:
    """Add leading zeros to the RID to keep à length of 4"""
    rid = str(ref_id)
    # Fill the rid with the needed number of zero
    if 4 - len(rid) > 0:
        zeros_to_add = 4 - len(rid)
        rid = "0" * zeros_to_add + rid
    return rid


def _compute_session_id(df: pd.DataFrame, csv_filename: str) -> pd.DataFrame:
    """Compute session ID from visit code.

    The CSV file name is used to determine in which column of
    the dataframe the visit code should be found.
    """
    visit_code_column = _get_visit_code_column_name(csv_filename)
    if visit_code_column not in df:
        raise ValueError(
            f"DataFrame does not contain a column named '{visit_code_column}', "
            "which is supposed to encode the visit code. The columns present in "
            f"the DataFrame are: {df.columns}."
        )
    return df.assign(
        session_id=lambda _df: _df[visit_code_column].apply(
            lambda x: _get_session_id_from_visit_code(x)
        )
    )


def _get_visit_code_column_name(csv_filename: str) -> str:
    """Return the name of the column containing the visit code."""
    if _is_a_visit_code_2_type(csv_filename):
        return "VISCODE2"
    if _is_a_time_point_type(csv_filename):
        return "Timepoint"
    return "VISCODE"


def _is_a_visit_code_2_type(csv_filename: str) -> bool:
    """If the csv file is among these files, then the visit code column is 'VISCODE2'."""
    return csv_filename in {
        "ADAS_ADNIGO2.csv",
        "DXSUM_PDXCONV.csv",
        "DXSUM_PDXCONV_ADNIALL.csv",
        "CDR.csv",
        "NEUROBAT.csv",
        "GDSCALE.csv",
        "MODHACH.csv",
        "MOCA.csv",
        "NPIQ.csv",
        "MEDHIST.csv",
        "VITALS.csv",
        "UWNPSYCHSUM_03_07_19.csv",
        "UWNPSYCHSUM_03_26_20.csv",
        "ECOGPT.csv",
        "ECOGSP.csv",
        "FCI.csv",
        "CCI.csv",
        "NPIQ.csv",
        "NPI.csv",
    }


def _is_a_time_point_type(csv_filename: str) -> bool:
    """If the csv file is among these files, then the visit code column is 'Timepoint'."""
    return csv_filename in {
        "BHR_EVERYDAY_COGNITION.csv",
        "BHR_BASELINE_QUESTIONNAIRE.csv",
        "BHR_LONGITUDINAL_QUESTIONNAIRE.csv",
    }


def _get_session_id_from_visit_code(visit_code: str) -> Optional[str]:
    """Checks that the visit code found in the data is supported by
    the converter utility function `viscode_to_session` before using it.
    There are a few special cases that are handled here.
    """
    from clinica.iotools.converter_utils import viscode_to_session

    if _is_visit_code_not_supported(visit_code):
        return None
    if visit_code == "sc":
        return "sc"
    return viscode_to_session(visit_code)


def _is_visit_code_not_supported(visit_code: str) -> bool:
    """Return True is the visit code is not supported by the converter.

    There are a few known values like "f" or "uns1" which are present in
    ADNI data that are not supported for a mapping to a session ID.
    """
    unsupported_values = {"f", "uns1"}
    return pd.isnull(visit_code) or visit_code in unsupported_values


def create_adni_sessions_dict(
    bids_ids: list[str],
    clinical_specifications_folder: Path,
    clinical_data_dir: Path,
    bids_subjs_paths: list[Path],
):
    """Extract all the data required for the sessions files and organize them in a dictionary.

    Args:
        bids_ids: list of subject IDs to process
        clinical_specifications_folder: path to the specifications for converting the clinical data
        clinical_data_dir: path to the clinical data folder
        bids_subjs_paths: a list with the path to all the BIDS subjects
    """
    from clinica.utils.stream import cprint

    df_sessions = pd.read_csv(clinical_specifications_folder / "sessions.tsv", sep="\t")
    files = list(df_sessions["ADNI location"].unique())
    files = [x for x in files if not pd.isnull(x)]
    df_subj_session = pd.DataFrame()
    # write line to get field_bids = sessions['BIDS CLINICA'] without the null values

    # Iterate over the metadata files

    for location in files:
        location = location.split("/")[0]
        try:
            df_file = load_clinical_csv(clinical_data_dir, location.split(".")[0])
        except IOError:
            continue
        df_filtered = _filter_subj_bids(df_file, location, bids_ids).copy()
        if not df_filtered.empty:
            df_filtered = _compute_session_id(df_filtered, location)
            # Filter rows with invalid session IDs.
            df_filtered.dropna(subset="session_id", inplace=True)

            if location == "ADNIMERGE.csv":
                df_filtered["AGE"] = df_filtered.apply(lambda x: _update_age(x), axis=1)
            df_subj_session = _update_sessions_df(
                df_subj_session, df_filtered, df_sessions, location
            )
        else:
            cprint(
                f"Clinical dataframe extracted from {location} is empty after filtering.",
                lvl="debug",
            )
            dict_column_correspondence = dict(
                zip(df_sessions["ADNI"], df_sessions["BIDS CLINICA"])
            )
            df_filtered.rename(columns=dict_column_correspondence, inplace=True)
            df_filtered = df_filtered.loc[
                :, (~df_filtered.columns.isin(df_subj_session.columns))
            ]
            df_subj_session = pd.concat([df_subj_session, df_filtered], axis=1)
    if df_subj_session.empty:
        cprint(
            "Empty dataset detected. Clinical data cannot be extracted.", lvl="warning"
        )
        return
    # Nv/None refer to sessions whose session is undefined. "sc" is the screening session with unreliable (incomplete)
    # data.
    df_subj_session = df_subj_session[
        df_subj_session.session_id.str.contains(r"ses-M\d+")
    ]
    _write_adni_sessions_tsv(df_subj_session, bids_subjs_paths)


def _update_sessions_df(
    df_subj_session: pd.DataFrame,
    df_filtered: pd.DataFrame,
    df_sessions: pd.DataFrame,
    location: str,
) -> pd.DataFrame:
    """Update the sessions dataframe with data of current subject.

    Args:
        df_subj_session: dataframe containing aggregate sessions
        df_filtered: dataframe with current subject sessions data
        df_sessions: dataframe with the the metadata information
        location: the clinica data filename
    """
    df_columns_to_add = df_sessions[
        df_sessions["ADNI location"].str.contains(location, na=False)
    ][["BIDS CLINICA", "ADNI"]]

    df_columns_to_add = df_columns_to_add[
        (df_columns_to_add["ADNI"].isin(df_filtered.columns))
        & (~df_columns_to_add["BIDS CLINICA"].isin(df_subj_session.columns))
    ]
    df_temp = df_filtered[
        ["RID", "session_id"] + list(df_columns_to_add["ADNI"])
    ].copy()
    df_temp.columns = ["RID", "session_id"] + list(df_columns_to_add["BIDS CLINICA"])

    # if error in adni data (duplicate session id), keep only the first row
    df_temp.drop_duplicates(subset=["RID", "session_id"], keep="first", inplace=True)
    try:
        df_temp["diagnosis"] = df_temp["diagnosis"].apply(
            lambda x: _convert_diagnosis_code(x)
        )
    except KeyError:
        pass
    if df_subj_session.empty:
        df_subj_session = df_temp
    else:
        df_subj_session = pd.merge(
            df_temp, df_subj_session, on=["RID", "session_id"], how="outer"
        )
    return df_subj_session


def _convert_diagnosis_code(diagnosis_code: str) -> str:
    """Convert the string field DX contained in ADNIMERGE.csv into a code that identify the diagnosis.

    Args:
        diagnosis_code: a string that represents the diagnosis

    Returns:
        A code that identify a diagnosis
    """
    diagnosis = {"CN": "CN", "MCI": "MCI", "Dementia": "AD"}

    if pd.isna(diagnosis_code):
        return diagnosis_code
    try:
        return diagnosis[diagnosis_code]
    except KeyError:
        raise KeyError(f"Unknown diagnosis code {diagnosis_code}.")


def create_adni_scans_files(conversion_path: Path, bids_subjs_paths: list[Path]):
    """Create scans.tsv files for ADNI.

    Args:
        conversion_path: path to the conversion_info folder
        bids_subjs_paths: list of bids subject paths
    """
    from os import path

    from clinica.iotools.bids_utils import StudyName, bids_id_factory
    from clinica.utils.stream import cprint

    scans_fields_bids = ["filename", "scan_id", "mri_field"]

    conversion_versions = [
        folder.name
        for folder in conversion_path.iterdir()
        if folder.name[0] == "v" and (conversion_path / folder).is_dir()
    ]
    conversion_versions = sorted(
        conversion_versions, key=lambda x: int(x.split("v")[1])
    )
    older_version = conversion_versions[-1]
    converted_dict = dict()
    for tsv_path in (conversion_path / older_version).iterdir():
        modality = tsv_path.name.split("_paths")[0]
        df = pd.read_csv(conversion_path / older_version / tsv_path, sep="\t")
        df.set_index(["Subject_ID", "VISCODE"], inplace=True, drop=True)
        converted_dict[modality] = df

    for bids_subject_path in bids_subjs_paths:
        # Create the file
        bids_id = bids_subject_path.resolve().name
        subject_id = bids_id_factory(StudyName.ADNI)(bids_id).to_original_study_id()
        for session_path in bids_subject_path.glob("ses-*"):
            viscode = _session_label_to_viscode(session_path.name[4::])
            tsv_name = f"{bids_id}_{session_path.name}_scans.tsv"
            (session_path / tsv_name).unlink(missing_ok=True)
            scans_tsv = open(session_path / tsv_name, "a")
            scans_df = pd.DataFrame(columns=scans_fields_bids)
            scans_df.to_csv(scans_tsv, sep="\t", index=False, encoding="utf-8")

            # Extract modalities available for each subject
            for mod in session_path.glob("*"):
                for file in mod.glob("*"):
                    scans_df = pd.DataFrame(index=[0], columns=scans_fields_bids)
                    scans_df["filename"] = path.join(mod.name, file.name)
                    converted_mod = _find_conversion_mod(file.name)
                    conversion_df = converted_dict[converted_mod]
                    try:
                        scan_id = conversion_df.loc[(subject_id, viscode), "Image_ID"]
                        scans_df["scan_id"] = scan_id
                        if "Field_Strength" in conversion_df.columns.values:
                            field_strength = conversion_df.loc[
                                (subject_id, viscode), "Field_Strength"
                            ]
                            scans_df["mri_field"] = field_strength
                    except KeyError:
                        cprint(
                            msg=f"No information found for file {file.name}",
                            lvl="warning",
                        )
                    scans_df = scans_df.fillna("n/a")
                    scans_df.to_csv(
                        scans_tsv, header=False, sep="\t", index=False, encoding="utf-8"
                    )


def _find_conversion_mod(file_name: str) -> str:
    from clinica.utils.pet import Tracer

    suffix = file_name.split("_")[-1].split(".")[0]

    if suffix == "T1w":
        return "t1"
    if suffix == "FLAIR":
        return "flair"
    if suffix == "dwi":
        return "dwi"
    if suffix == "bold":
        return "fmri"
    if suffix in (
        "fmap",
        "ph",
        "e1",
        "e2",
        "phase1",
        "phase2",
        "phasediff",
        "magnitude1",
        "magnitude2",
    ):
        return "fmap"
    if suffix == "pet":
        tracer = file_name.split("trc-")[1].split("_")[0]
        if tracer in (Tracer.AV45, Tracer.FBB):
            return "amyloid_pet"
        return f"{tracer}_pet"
    raise ValueError(f"Conversion modality could not be found for file {file_name}")


def paths_to_bids(
    images: pd.DataFrame,
    bids_dir: Path,
    modality: ADNIModalityConverter,
    mod_to_update: bool = False,
    n_procs: Optional[int] = 1,
) -> List[Path]:
    """Images in the list are converted and copied to directory in BIDS format.

    Parameters
    ----------
    images : pd.DataFrame
        List of images metadata and paths.

    bids_dir : Path
        The path to the output BIDS directory.

    modality : str
        Imaging modality.

    mod_to_update : bool
        If True, pre-existing images in the BIDS directory will be erased
        and extracted again.

    n_procs : int, optional
        The number of processes to use for conversion.
        Default=1.

    Returns
    -------
    output_file_treated : list of paths
        List of path to image files created.
        This list contains None values for files where the
        conversion wasn't successful.
    """
    from functools import partial
    from multiprocessing import Pool

    images_list = list([data for _, data in images.iterrows()])
    create_file_ = partial(
        _create_file,
        modality=modality,
        bids_dir=bids_dir,
        mod_to_update=mod_to_update,
    )
    # If n_procs==1 do not rely on a Process Pool to enable classical debugging
    if n_procs == 1:
        return [create_file_(image) for image in images_list]
    with Pool(processes=n_procs) as pool:
        output_file_treated = pool.map(create_file_, images_list)
    return output_file_treated


def _get_images_with_suffix(
    folder: Path,
    suffixes: Iterable[str],
) -> Iterable[Path]:
    import re

    if suffixes:
        rgx = re.compile(r"|".join(suffixes))
        images = list(filter(rgx.search, [f.name for f in folder.iterdir()]))
        return [folder / image for image in images]
    return []


def _remove_existing_images_if_necessary(
    folder: Path,
    suffixes: Iterable[str],
    mod_to_update: bool = False,
) -> bool:
    """Returns whether it was possible to delete the existing images or not."""
    from clinica.utils.stream import cprint

    images_to_remove = _get_images_with_suffix(folder, suffixes)
    if not images_to_remove:
        return True
    elif mod_to_update:
        cprint(f"Removing old images : {images_to_remove}", lvl="info")
        for path_to_unlink in images_to_remove:
            (path_to_unlink).unlink()
        return True
    cprint(
        f"There already exist images : {images_to_remove}. "
        "The parameter 'mod_to_update' is set to False so that "
        "they cannot be overwritten.",
        lvl="warning",
    )
    return False


def _remove_files_with_unsupported_suffixes(
    folder: Path, suffixes: Iterable[str]
) -> Iterable[Path]:
    from clinica.utils.stream import cprint

    invalid_suffix_images = _get_images_with_suffix(folder, suffixes)
    for image_to_remove in invalid_suffix_images:
        cprint(
            f"Image with bad suffix {', '.join(suffixes)} was generated by dcm2nix. These are not"
            f"supported by Clinica so the image will NOT be converted.",
            lvl="warning",
        )
        cprint(f"Removing image {folder / image_to_remove}", lvl="info")
        (folder / image_to_remove).unlink()
    return [folder / image for image in invalid_suffix_images]


def _create_file(
    image: pd.Series,
    modality: ADNIModalityConverter,
    bids_dir: Path,
    mod_to_update: bool,
) -> Optional[Path]:
    """Creates an image file at the corresponding output folder.

    The image file is the result of image conversion (DICOM to NIFTI)
    and centering, as specified for each modality

    Parameters
    ----------
    image : pd.Series
        The image metadata.

    modality : str
        Imaging modality.

    bids_dir : Path
        The path to the output BIDS directory.

    mod_to_update : bool
        If True, pre-existing images in the BIDS directory will be
        erased and extracted again.

    Returns
    -------
    output_image : Path
        The path to the output image created.
        If the conversion wasn't successful, this function returns None.
    """
    import os
    import re
    import shutil

    import numpy as np

    from clinica.cmdline import setup_clinica_logging
    from clinica.iotools.bids_utils import StudyName, bids_id_factory, run_dcm2niix
    from clinica.iotools.converter_utils import viscode_to_session
    from clinica.iotools.utils.data_handling import center_nifti_origin
    from clinica.utils.stream import cprint, log_and_raise, log_and_warn

    # This function is executed in a multiprocessing context
    # such that we need to re-configure the clinica logger in the child processes.
    # Note that logging messages could easily be lost (for example when logging
    # to a file from two different processes). A better solution would be to
    # implement a logging process consuming logging messages from a multiprocessing.Queue...
    setup_clinica_logging("INFO")

    subject = image.Subject_ID
    viscode = image.VISCODE

    image_tracer = None
    logging_header = f"[{modality.value}]"
    if modality == ADNIModalityConverter.PET_AV45:
        image_tracer = Tracer(image.Tracer)
        logging_header = f"[{modality.value} - {image_tracer.value}]"
    if not image.Path:
        # todo : check for message redundancy
        cprint(
            f"{logging_header} No path specified for {subject} in session {viscode}",
            lvl="info",
        )
        return None
    cprint(
        f"{logging_header} Processing subject {subject} in session {viscode}",
        lvl="info",
    )
    session = viscode_to_session(viscode)
    image_path = Path(image.Path)
    image_id = image.Image_ID
    # If the original image is a DICOM, check if contains two DICOM inside the same folder
    if image.Is_Dicom:
        image_path = _check_two_dcm_folder(image_path, bids_dir, image_id)
    bids_id = bids_id_factory(StudyName.ADNI).from_original_study_id(subject)
    output_path = bids_dir / bids_id / session / _get_output_path(modality)
    output_filename = (
        f"{bids_id}_{session}{_get_output_filename(modality, image_tracer)}"
    )
    output_path.mkdir(parents=True, exist_ok=True)

    # Check if old images exist and if updated mode is set to TRUE remove them
    if not _remove_existing_images_if_necessary(
        output_path,
        (_get_output_filename(modality, image_tracer), "magnitude", "phase"),
        mod_to_update,
    ):
        return None

    file_without_extension = output_path / output_filename
    output_image = file_without_extension.with_suffix(".nii.gz")

    if image.Is_Dicom:
        success = run_dcm2niix(
            input_dir=image_path
            if modality != ADNIModalityConverter.FMAP
            else Path(image_path).parent,
            output_dir=output_path,
            output_fmt=output_filename,
            compress=not _should_be_centered(modality),
            bids_sidecar=_write_json_sidecar(modality),
        )
        if not success:
            log_and_warn(
                f"{logging_header} Error converting image {image_path} for subject {subject} and session {session}",
                UserWarning,
            )

        # If "_t" - the trigger delay time - exists in dcm2niix output filename, we remove it
        for trigger_time in output_path.glob(f"{output_filename}_t[0-9]*"):
            res = re.search(r"_t\d+\.", str(trigger_time))
            no_trigger_time = str(trigger_time).replace(
                trigger_time[res.start() : res.end()], "."
            )
            os.rename(trigger_time, no_trigger_time)

        # Removing images with unsupported suffixes if generated by dcm2niix
        # todo : maybe later use output to remind user what failed
        _remove_files_with_unsupported_suffixes(
            output_path, ("ADC", "real", "imaginary")
        )

        # Conditions to check if output NIFTI files exists,
        # and, if DWI, if .bvec and .bval files are also present
        nifti_exists = (
            file_without_extension.with_suffix(".nii").is_file()
            or output_image.is_file()
        )
        if modality == ADNIModalityConverter.FMAP:
            nifti_exists = nifti_exists or any(
                [
                    file_without_extension.with_name(
                        f"{file_without_extension.name}_{suffix}.nii.gz"
                    ).is_file()
                    for suffix in {"ph", "e2", "e1", "e2_ph", "e1_ph"}
                ]
            )
        dwi_bvec_and_bval_exist = not (modality == ADNIModalityConverter.DWI) or (
            file_without_extension.with_suffix(".bvec").is_file()
            and file_without_extension.with_suffix(".bval").is_file()
        )

        # Check if conversion worked (output files exist)
        if not nifti_exists or not dwi_bvec_and_bval_exist:
            log_and_warn(
                f"{logging_header} Conversion with dcm2niix failed for subject {subject} and session {session}",
                UserWarning,
            )
            return np.nan

        # Case when JSON file was expected, but not generated by dcm2niix
        json_exists = file_without_extension.with_suffix(".json").exists()

        if modality == ADNIModalityConverter.FMAP:
            json_exists = json_exists or any(
                [
                    file_without_extension.with_name(
                        f"{file_without_extension.name}_{suffix}.json"
                    ).is_file()
                    for suffix in {"ph", "e2", "e1", "e2_ph", "e1_ph"}
                ]
            )

        if _write_json_sidecar(modality) and not json_exists:
            log_and_warn(
                f"{logging_header} JSON file not generated by dcm2niix for subject {subject} and session {session}",
                UserWarning,
            )
        if _should_be_centered(modality):
            try:
                output_image = center_nifti_origin(
                    file_without_extension.with_suffix(".nii"), output_image
                )
            except Exception as e:
                log_and_raise(str(e), ValueError)
            file_without_extension.with_suffix(".nii").unlink()
    else:
        if _should_be_centered(modality):
            try:
                output_image = center_nifti_origin(image_path, output_image)
            except Exception as e:
                log_and_raise(
                    (
                        f"For subject {subject} in session {session}, "
                        f"an error occurred whilst recentering Nifti image: {image_path}"
                        f"The error is: {e}"
                    ),
                    ValueError,
                )
        else:
            shutil.copy(image_path, output_image)

    # Check if there is still the folder tmp_dcm_folder and remove it
    _remove_tmp_dmc_folder(bids_dir, image_id)
    return output_image


def _get_output_path(modality: ADNIModalityConverter) -> str:
    if modality == ADNIModalityConverter.T1:
        return "anat"
    if modality == ADNIModalityConverter.DWI:
        return "dwi"
    if modality == ADNIModalityConverter.FLAIR:
        return "anat"
    if modality == ADNIModalityConverter.FMRI:
        return "func"
    if modality == ADNIModalityConverter.FMAP:
        return "fmap"
    if modality in (
        ADNIModalityConverter.PET_FDG,
        ADNIModalityConverter.PET_FDG_UNIFORM,
        ADNIModalityConverter.PET_PIB,
        ADNIModalityConverter.PET_AV45,
        ADNIModalityConverter.PET_TAU,
    ):
        return "pet"


def _get_output_filename(
    modality: ADNIModalityConverter, tracer: Optional[Tracer] = None
) -> str:
    if modality == ADNIModalityConverter.T1:
        return "_T1w"
    if modality == ADNIModalityConverter.DWI:
        return "_dwi"
    if modality == ADNIModalityConverter.FLAIR:
        return "_FLAIR"
    if modality == ADNIModalityConverter.FMRI:
        return "_task-rest_bold"
    if modality == ADNIModalityConverter.FMAP:
        return "_fmap"
    if modality == ADNIModalityConverter.PET_FDG:
        return f"_trc-{Tracer.FDG.value}_rec-{ReconstructionMethod.CO_REGISTERED_AVERAGED.value}_pet"
    if modality == ADNIModalityConverter.PET_FDG_UNIFORM:
        return f"_trc-{Tracer.FDG.value}_rec-{ReconstructionMethod.COREGISTERED_ISOTROPIC.value}_pet"
    if modality == ADNIModalityConverter.PET_PIB:
        return f"_trc-{Tracer.PIB.value}_pet"
    if modality == ADNIModalityConverter.PET_AV45:
        return f"_trc-{tracer.value}_pet"
    if modality == ADNIModalityConverter.PET_TAU:
        return f"_trc-{Tracer.AV1451.value}_pet"


def _should_be_centered(modality: ADNIModalityConverter) -> bool:
    if modality == ADNIModalityConverter.T1:
        return True
    if modality == ADNIModalityConverter.DWI:
        return False
    if modality == ADNIModalityConverter.FLAIR:
        return True
    if modality == ADNIModalityConverter.FMRI:
        return False
    if modality == ADNIModalityConverter.FMAP:
        return False
    if modality in (
        ADNIModalityConverter.PET_FDG,
        ADNIModalityConverter.PET_FDG_UNIFORM,
        ADNIModalityConverter.PET_PIB,
        ADNIModalityConverter.PET_AV45,
        ADNIModalityConverter.PET_TAU,
    ):
        return True


def _write_json_sidecar(modality: ADNIModalityConverter) -> bool:
    if modality == ADNIModalityConverter.T1:
        return False
    if modality == ADNIModalityConverter.DWI:
        return True
    if modality == ADNIModalityConverter.FLAIR:
        return True
    if modality == ADNIModalityConverter.FMRI:
        return True
    if modality == ADNIModalityConverter.FMAP:
        return True
    if modality in (
        ADNIModalityConverter.PET_FDG,
        ADNIModalityConverter.PET_FDG_UNIFORM,
        ADNIModalityConverter.PET_PIB,
        ADNIModalityConverter.PET_AV45,
        ADNIModalityConverter.PET_TAU,
    ):
        return False


def _session_label_to_viscode(session_name: str) -> str:
    """Replace the session name passed as input with the session label 'bl' or 'mXXX'.

    Parameters
    ----------
    session_name: str
        Name of the session (MXXX).

    Returns
    -------
    str:
        'bl' if is the baseline session or the original session name.
    """
    if session_name == "M000":
        return "bl"
    return f"m{(int(session_name[1:])):02d}"


def _check_two_dcm_folder(dicom_path: Path, bids_folder: Path, image_uid: str) -> Path:
    """[summary].

    Check if a folder contains more than one DICOM and if yes, copy the DICOM related to
    image id passed as parameter into a temporary folder called tmp_dicom_folder.

    Parameters
    ----------
    dicom_path : Path
        The path to the DICOM folder.

    bids_folder : Path
        The path to the BIDS folder where the dataset will be stored.

    image_uid : str
        The image id of the fMRI.

    Returns
    -------
    Path :
        The path to the original DICOM folder or the path to a temporary folder called
        tmp_dicom_folder where only the DICOM to convert is copied.
    """
    import shutil
    from shutil import copy

    dest_path = bids_folder / f"tmp_dcm_folder_{image_uid.strip(' ')}"
    # Check if there are dicom files inside the folder not belonging to the desired image
    dicom_list = [f for f in dicom_path.glob("*.dcm")]
    image_list = [f for f in dicom_path.glob(f"*{image_uid}.dcm")]
    if len(dicom_list) != len(image_list):
        # Remove the precedent tmp_dcm_folder if present.
        if dest_path.exists():
            shutil.rmtree(dest_path)
        dest_path.mkdir(parents=True)
        for d in dicom_path.glob(f"*{str(image_uid)}.dcm*"):
            copy(d, dest_path)
        return dest_path
    return dicom_path


def _remove_tmp_dmc_folder(bids_dir: Path, image_id: str):
    """Remove the folder tmp_dmc_folder created by the method check_two_dcm_folder (if existing).

    Args:
        bids_dir: path to the BIDS directory
        image_id:
    """
    from shutil import rmtree

    tmp_dcm_folder_path = bids_dir / f"tmp_dcm_folder_{image_id.strip(' ')}"
    if tmp_dcm_folder_path.exists():
        rmtree(tmp_dcm_folder_path)


def load_clinical_csv(clinical_dir: Path, filename: str) -> pd.DataFrame:
    """Load the clinical csv from ADNI. This function is able to find the csv in the
    different known format available, the old format with just the name, and the new
    format with the name and the date of download.

    Parameters
    ----------
    clinical_dir: str
        Directory containing the csv.

    filename: str
        name of the file without the suffix.

    Returns
    -------
    pd.DataFrame:
        Dataframe corresponding to the filename.
    """
    import re

    pattern = filename + r"(_\d{1,2}[A-Za-z]{3}\d{4})?.csv"
    files_matching_pattern = [
        f for f in clinical_dir.rglob("[!.]*.csv") if re.search(pattern, f.name)
    ]
    if len(files_matching_pattern) != 1:
        raise IOError(
            f"Expecting to find exactly one file in folder {clinical_dir} "
            f"matching pattern {pattern}. {len(files_matching_pattern)} "
            f"files were found instead : \n{'- '.join(str(files_matching_pattern))}"
        )
    try:
        return pd.read_csv(files_matching_pattern[0], sep=",", low_memory=False)
    except Exception:
        raise ValueError(
            f"File {str(files_matching_pattern[0])} was found but could not "
            "be loaded as a DataFrame. Please check your data."
        )



from typing import List, Optional

import click

import clinica.pydra.engine_utils as pydra_utils
import clinica.pydra.t1_volume.tissue_segmentation.pipeline as pydra_t1vol
from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "pydra-t1-volume-tissue-segmentation"


@clinica_pipeline
@click.command(name=pipeline_name, hidden=True)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.yes
@option.global_option_group
@option.n_procs
@cli_param.option_group.advanced_pipeline_options
@cli_param.option.tissue_classes
@cli_param.option.dartel_tissues
@cli_param.option.tissue_probability_maps
@cli_param.option.dont_save_warped_unmodulated
@cli_param.option.save_warped_modulated
def cli(
    bids_directory: str,
    caps_directory: str,
    tissue_classes: List[int] = (1, 2, 3),
    dartel_tissues: List[int] = (1, 2, 3),
    tissue_probability_maps: Optional[str] = None,
    dont_save_warped_unmodulated: bool = False,
    save_warped_modulated: bool = False,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
    yes: bool = False,
) -> None:
    """Affine registration of Flair images to the MNI standard space (Pydra engine)."""

    parameters = {
        "tissue_classes": tissue_classes,
        "dartel_tissues": dartel_tissues,
        "tissue_probability_maps": tissue_probability_maps,
        "save_warped_unmodulated": not dont_save_warped_unmodulated,
        "save_warped_modulated": save_warped_modulated,
        "skip_question": yes,
    }

    t1_volume_tissue_segmentation_pipeline = pydra_t1vol.t1volume_tissue_segmentation(
        name="t1-volume-tissue-segmentation-pydra",
        input_dir=bids_directory,
        output_dir=caps_directory,
        parameters=parameters,
    )
    pydra_utils.run(t1_volume_tissue_segmentation_pipeline, n_procs=n_procs)


if __name__ == "__main__":
    cli()
